# Copyright (c) 2021-2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Parallax correction.

Routines related to parallax correction using datasets involving height, such
as cloud top height.

The geolocation of (geostationary) satellite imagery is calculated by
agencies or in satpy readers with the assumption of a clear view from
the satellite to the geoid.  When a cloud blocks the view of the Earth
surface or the surface is above sea level, the geolocation is not accurate
for the cloud or mountain top.  This module contains routines to correct
imagery such that pixels are shifted or interpolated to correct for this
parallax effect.

Parallax correction is currently only supported for (cloud top) height
that arrives on an :class:`~pyresample.geometry.AreaDefinition`, such
as is standard for geostationary satellites.  Parallax correction with
data described by a :class:`~pyresample.geometry.SwathDefinition`,
such as is common for polar satellites, is not (yet) supported.

See also the :doc:`../modifiers` page in the documentation for an introduction to
parallax correction as a modifier in Satpy.
"""

import datetime
import inspect
import logging
import warnings

import dask.array as da
import numpy as np
import xarray as xr
from pyorbital.orbital import A as EARTH_RADIUS
from pyorbital.orbital import get_observer_look
from pyproj import Geod
from pyresample.bucket import BucketResampler
from pyresample.geometry import SwathDefinition

from satpy.modifiers import ModifierBase
from satpy.resample.base import resample_dataset
from satpy.utils import get_satpos, lonlat2xyz, xyz2lonlat

logger = logging.getLogger(__name__)


class MissingHeightError(ValueError):
    """Raised when heights do not overlap with area to be corrected."""


class IncompleteHeightWarning(UserWarning):
    """Raised when heights only partially overlap with area to be corrected."""


def get_parallax_corrected_lonlats(sat_lon, sat_lat, sat_alt, lon, lat, height):
    """Calculate parallax corrected lon/lats.

    Satellite geolocation generally assumes an unobstructed view of a smooth
    Earth surface.  In reality, this view may be obstructed by clouds or
    mountains.

    If the view of a pixel at location (lat, lon) is blocked by a cloud
    at height h, this function calculates the (lat, lon) coordinates
    of the cloud above/in front of the invisible surface.

    For scenes that are only partly cloudy, the user might set the cloud top
    height for clear-sky pixels to NaN.  This function will return a corrected
    lat/lon as NaN as well.  The user can use the original lat/lon for those
    pixels or use the higher level :class:`ParallaxCorrection` class.

    This function assumes a spherical Earth.

    .. note::

        Be careful with units!  This code expects ``sat_alt`` and
        ``height`` to be in meter above the Earth's surface.  You may
        have to convert your input correspondingly.  Cloud Top Height
        is usually reported in meters above the Earth's surface, rarely
        in km.  Satellite altitude may be reported in either m or km, but
        orbital parameters are usually in relation to the Earth's centre.
        The Earth radius from pyresample is reported in km.

    Args:
        sat_lon (numbers.Number): Satellite longitude in geodetic coordinates [degrees]
        sat_lat (numbers.Number): Satellite latitude in geodetic coordinates [degrees]
        sat_alt (numbers.Number): Satellite altitude above the Earth surface [m]
        lon (dask.array.Array or numbers.Number): Longitudes of pixel or pixels to be corrected,
            in geodetic coordinates [degrees]
        lat (dask.array.Array or numbers.Number): Latitudes of pixel/pixels to be corrected, in
            geodetic coordinates [degrees]
        height (dask.array.Array or numbers.Number): Heights of pixels on which the correction
            will be based.  Typically this is the cloud top height. [m]

    Returns:
        tuple[float, float]: Corrected geolocation
            Corrected geolocation ``(lon, lat)`` in geodetic coordinates for
            the pixel(s) to be corrected. [degrees]
    """
    elevation = _get_satellite_elevation(sat_lon, sat_lat, sat_alt, lon, lat)
    parallax_distance = _calculate_slant_cloud_distance(height, elevation)
    shifted_xyz = _get_parallax_shift_xyz(
            sat_lon, sat_lat, sat_alt, lon, lat, parallax_distance)

    return xyz2lonlat(
        shifted_xyz[..., 0], shifted_xyz[..., 1], shifted_xyz[..., 2])


def get_surface_parallax_displacement(
        sat_lon, sat_lat, sat_alt, lon, lat, height):
    """Calculate surface parallax displacement.

    Calculate the displacement due to parallax error.  Input parameters are
    identical to :func:`get_parallax_corrected_lonlats`.

    Returns:
        (numbers.Number, dask.array.Array): parallax displacement in meters
    """
    (corr_lon, corr_lat) = get_parallax_corrected_lonlats(sat_lon, sat_lat, sat_alt, lon, lat, height)
    # Get parallax displacement
    geod = Geod(ellps="sphere")
    _, _, parallax_dist = geod.inv(corr_lon, corr_lat, lon, lat)
    return parallax_dist


def _get_parallax_shift_xyz(sat_lon, sat_lat, sat_alt, lon, lat, parallax_distance):
    """Calculate the parallax shift in cartesian coordinates.

    From satellite position and cloud position, get the parallax shift in
    cartesian coordinates:

    Args:
        sat_lon (numbers.Number): Satellite longitude in geodetic coordinates [degrees]
        sat_lat (numbers.Number): Satellite latitude in geodetic coordinates [degrees]
        sat_alt (numbers.Number): Satellite altitude above the Earth surface [m]
        lon (dask.array.Array or numbers.Number): Longitudes of pixel or pixels to be corrected,
            in geodetic coordinates [degrees]
        lat (dask.array.Array or numbers.Number): Latitudes of pixel/pixels to be corrected, in
            geodetic coordinates [degrees]
        parallax_distance (dask.array.Array or numbers.Number): Cloud to ground distance with parallax
            effect [m].

    Returns:
        Parallax shift in cartesian coordinates in meter.
    """
    sat_xyz = np.hstack(lonlat2xyz(sat_lon, sat_lat)) * (sat_alt + EARTH_RADIUS * 1e3)
    cth_xyz = np.stack(lonlat2xyz(lon, lat), axis=-1) * EARTH_RADIUS*1e3  # km → m
    delta_xyz = cth_xyz - sat_xyz
    sat_distance = np.sqrt((delta_xyz*delta_xyz).sum(axis=-1))
    dist_shape = delta_xyz.shape[:-1] + (1,)  # force correct array broadcasting
    return cth_xyz - delta_xyz*(parallax_distance/sat_distance).reshape(dist_shape)


def _get_satellite_elevation(sat_lon, sat_lat, sat_alt, lon, lat):
    """Get satellite elevation.

    Get the satellite elevation from satellite lon/lat/alt for positions
    lon/lat.
    """
    placeholder_date = datetime.datetime(2000, 1, 1)  # no impact on get_observer_look?
    (_, elevation) = get_observer_look(
            sat_lon, sat_lat, sat_alt/1e3,  # m → km (wanted by get_observer_look)
            placeholder_date, lon, lat, 0)
    return elevation


def _calculate_slant_cloud_distance(height, elevation):
    """Calculate slant cloud to ground distance.

    From (cloud top) height and satellite elevation, calculate the
    slant cloud-to-ground distance along the line of sight of the satellite.
    """
    if np.isscalar(elevation) and elevation == 0:
        raise NotImplementedError(
                "Parallax correction not implemented for "
                "satellite elevation 0")
    if np.isscalar(elevation) and elevation < 0:
        raise ValueError(
                "Satellite is below the horizon.  Cannot calculate parallax "
                "correction.")
    return height / np.sin(np.deg2rad(elevation))


class ParallaxCorrection:
    """Parallax correction calculations.

    This class contains higher-level functionality to wrap the parallax
    correction calculations in :func:`get_parallax_corrected_lonlats`.  The class is
    initialised using a base area, which is the area for which a corrected
    geolocation will be calculated.  The resulting object is a callable.
    Calling the object with an array of (cloud top) heights returns a
    :class:`~pyresample.geometry.SwathDefinition` describing the new ,
    corrected geolocation.  The cloud top height should cover at least the
    area for which the corrected geolocation will be calculated.

    Note that the ``ctth`` dataset must contain satellite location
    metadata, such as set in the ``orbital_parameters`` dataset attribute
    that is set by many Satpy readers.  It is essential that the datasets to be
    corrected are coming from the same platform as the provided cloud top
    height.

    A note on the algorithm and the implementation.  Parallax correction
    is inherently an inverse problem.  The reported geolocation in
    satellite data files is the true location plus the parallax error.
    Therefore, this class first calculates the true geolocation (using
    :func:`get_parallax_corrected_lonlats`), which gives a shifted longitude and
    shifted latitude on an irregular grid.  The difference between
    the original and the shifted grid is the parallax error or shift.
    The magnitude of this error can be estimated with
    :func:`get_surface_parallax_displacement`.
    With this difference, we need to invert the parallax correction to
    calculate the corrected geolocation.  Due to parallax correction,
    high clouds shift a lot, low clouds shift a little, and cloud-free
    pixels shift not at all.  The shift may result in zero, one,
    two, or more source pixel onto a destination pixel.  Physically,
    this corresponds to the situation where a narrow but high cloud is
    viewed at a large angle.  The cloud may occupy two or more pixels when
    viewed at a large angle, but only one when viewed straight from above.
    To accurately reproduce this perspective, the parallax correction uses
    the :class:`~pyresample.bucket.BucketResampler` class, specifically
    the :meth:`~pyresample.bucket.BucketResampler.get_abs_max` method, to
    retain only the largest absolute shift (corresponding to the highest
    cloud) within each pixel.  Any other resampling method at this step
    would yield incorrect results.  When cloud moves over clear-sky, the
    clear-sky pixel is unshifted and the shift is located exactly in the
    centre of the grid box, so nearest-neighbour resampling would lead to
    such shifts being deselected.  Other resampling methods would average
    large shifts with small shifts, leading to unpredictable results.
    Now the reprojected shifts can be applied to the original lat/lon,
    returning a new :class:`~pyresample.geometry.SwathDefinition`.
    This is is the object returned by :meth:`corrected_area`.

    This procedure can be configured as a modifier using the
    :class:`ParallaxCorrectionModifier` class.  However, the modifier can only
    be applied to one dataset at the time, which may not provide optimal
    performance, although dask should reuse identical calculations between
    multiple channels.

    """

    def __init__(self, base_area,
                 debug_mode=False):
        """Initialise parallax correction class.

        Args:
            base_area (pyresample.geometry.AreaDefinition): Area for which calculated
                geolocation will be calculated.
            debug_mode (bool): Store diagnostic information in
                self.diagnostics.  This attribute always apply to the most
                recently applied operation only.
        """
        self.base_area = base_area
        self.debug_mode = debug_mode
        self.diagnostics = {}

    def __call__(self, cth_dataset, **kwargs):  # noqa: D417
        """Apply parallax correction to dataset.

        Args:
            cth_dataset: Dataset containing cloud top heights (or other heights
                to be corrected).

        Returns:
            :class:'~pyresample.geometry.SwathDefinition`: Swathdefinition with corrected
                lat/lon coordinates.
        """
        self.diagnostics.clear()
        return self.corrected_area(cth_dataset, **kwargs)

    def corrected_area(self, cth_dataset,
                       cth_resampler="nearest",
                       cth_radius_of_influence=50000,
                       lonlat_chunks=1024):
        """Return the parallax corrected SwathDefinition.

        Using the cloud top heights provided in ``cth_dataset``, calculate the
        :class:`pyresample.geometry.SwathDefinition` that estimates the
        geolocation for each pixel if it had been viewed from straight above
        (without parallax error).  The cloud top height will first be resampled
        onto the area passed upon class initialisation in :meth:`__init__`.
        Pixels that are invisible after parallax correction are not retained
        but get geolocation NaN.

        Args:
            cth_dataset (xarray.DataArray): Cloud top height in
                meters.  The variable attributes must contain an ``area``
                attribute describing the geolocation in a pyresample-aware way,
                and they must contain satellite orbital parameters.  The
                dimensions must be ``(y, x)``.  For best performance, this
                should be a dask-based :class:`~xarray.DataArray`.
            cth_resampler (str, Optional): Resampler to use when resampling the
                (cloud top) height to the base area.  Defaults to "nearest".
            cth_radius_of_influence (numbers.Number, Optional): Radius of influence to use when
                resampling the (cloud top) height to the base area.  Defaults
                to 50000.
            lonlat_chunks (int, Optional): Chunking to use when calculating lon/lats.
                Probably the default (1024) should be fine.

        Returns:
            :class:`~pyresample.geometry.SwathDefinition` describing parallax
            corrected geolocation.
        """
        logger.debug("Calculating parallax correction using heights from "
                     f"{cth_dataset.attrs.get('name', cth_dataset.name)!s}, "
                     f"with base area {self.base_area.description!s}.")
        (sat_lon, sat_lat, sat_alt_m) = _get_satpos_from_cth(cth_dataset)
        self._check_overlap(cth_dataset)

        cth_dataset = self._prepare_cth_dataset(
                cth_dataset, resampler=cth_resampler,
                radius_of_influence=cth_radius_of_influence,
                lonlat_chunks=lonlat_chunks)

        (base_lon, base_lat) = self.base_area.get_lonlats(chunks=lonlat_chunks)
        # calculate the shift/error due to the parallax effect
        (corrected_lon, corrected_lat) = get_parallax_corrected_lonlats(
                sat_lon, sat_lat, sat_alt_m,
                base_lon, base_lat, cth_dataset.data)

        shifted_area = self._get_swathdef_from_lon_lat(corrected_lon, corrected_lat)

        # But we are not actually moving pixels, rather we want a
        # coordinate transformation. With this transformation we approximately
        # invert the pixel coordinate transformation, giving the lon and lat
        # where we should retrieve a value for a given pixel.
        (proj_lon, proj_lat) = self._get_corrected_lon_lat(
                base_lon, base_lat, shifted_area)

        return self._get_swathdef_from_lon_lat(proj_lon, proj_lat)

    @staticmethod
    def _get_swathdef_from_lon_lat(lon, lat):
        """Return a SwathDefinition from lon/lat.

        Turn ndarrays describing lon/lat into xarray with dimensions y, x, then
        use these to create a :class:`~pyresample.geometry.SwathDefinition`.
        """
        # lons and lats passed to SwathDefinition must be data-arrays with
        # dimensions,  see https://github.com/pytroll/satpy/issues/1434
        # and https://github.com/pytroll/satpy/issues/1997
        return SwathDefinition(
            xr.DataArray(lon, dims=("y", "x")),
            xr.DataArray(lat, dims=("y", "x")))

    def _prepare_cth_dataset(
            self, cth_dataset, resampler="nearest", radius_of_influence=50000,
            lonlat_chunks=1024):
        """Prepare CTH dataset.

        Set cloud top height to zero wherever lat/lon are valid but CTH is
        undefined.  Then resample onto the base area.
        """
        # for calculating the parallax effect, set cth to 0 where it is
        # undefined, unless pixels have no valid lat/lon
        # NB: 0 may be below the surface... could be a problem for high
        # resolution imagery in mountainous or high elevation terrain
        # NB: how tolerant of xarray & dask is this?
        resampled_cth_dataset = resample_dataset(
                cth_dataset, self.base_area, resampler=resampler,
                radius_of_influence=radius_of_influence)
        (pixel_lon, pixel_lat) = resampled_cth_dataset.attrs["area"].get_lonlats(
                chunks=lonlat_chunks)
        masked_resampled_cth_dataset = resampled_cth_dataset.where(
                np.isfinite(pixel_lon) & np.isfinite(pixel_lat))
        masked_resampled_cth_dataset = masked_resampled_cth_dataset.where(
                masked_resampled_cth_dataset.notnull(), 0)
        return masked_resampled_cth_dataset

    def _check_overlap(self, cth_dataset):
        """Ensure cth_dataset is usable for parallax correction.

        Checks the coverage of ``cth_dataset`` compared to the ``base_area``.  If
        the entirety of ``base_area`` is covered by ``cth_dataset``, do
        nothing.  If only part of ``base_area`` is covered by ``cth_dataset``,
        raise a `IncompleteHeightWarning`.  If none of ``base_area`` is covered
        by ``cth_dataset``, raise a `MissingHeightError`.
        """
        warnings.warn(
            "Overlap checking not implemented. Waiting for "
            "fix for https://github.com/pytroll/pyresample/issues/329",
            stacklevel=3
        )

    def _get_corrected_lon_lat(self, base_lon, base_lat, shifted_area):
        """Calculate the corrected lon/lat based from the shifted area.

        After calculating the shifted area based on
        :func:`get_parallax_corrected_lonlats`,
        we invert the parallax error and estimate where those pixels came from.
        For details on the algorithm, see the class docstring.
        """
        (corrected_lon, corrected_lat) = shifted_area.get_lonlats(chunks=1024)
        lon_diff = corrected_lon - base_lon
        lat_diff = corrected_lat - base_lat
        # We use the bucket resampler here, because parallax correction
        # inevitably means there will be 2 source pixels ending up in the same
        # destination pixel.  We want to choose the biggest shift (max abs in
        # lat_diff and lon_diff), because the biggest shift corresponds to the
        # highest clouds, and if we move a 10 km cloud over a 2 km one, we
        # should retain the 10 km.
        #
        # some things to keep in mind:
        # - even with a constant cloud height, 3 source pixels may end up in
        #   the same destination pixel, because pixels get larger in the
        #   direction of the satellite.  This means clouds may shrink as they
        #   approach the satellite.
        # - the x-shift is a function of y and the y-shift is a function of x,
        #   so a cloud that was rectangular at the start may no longer be
        #   rectangular at the end
        bur = BucketResampler(self.base_area,
                              da.array(corrected_lon), da.array(corrected_lat))
        inv_lat_diff = bur.get_abs_max(lat_diff)
        inv_lon_diff = bur.get_abs_max(lon_diff)

        inv_lon = base_lon - inv_lon_diff
        inv_lat = base_lat - inv_lat_diff
        if self.debug_mode:
            self.diagnostics["corrected_lon"] = corrected_lon
            self.diagnostics["corrected_lat"] = corrected_lat
            self.diagnostics["inv_lon"] = inv_lon
            self.diagnostics["inv_lat"] = inv_lat
            self.diagnostics["inv_lon_diff"] = inv_lon_diff
            self.diagnostics["inv_lat_diff"] = inv_lat_diff
            self.diagnostics["base_lon"] = base_lon
            self.diagnostics["base_lat"] = base_lat
            self.diagnostics["lon_diff"] = lon_diff
            self.diagnostics["lat_diff"] = lat_diff
            self.diagnostics["shifted_area"] = shifted_area
            self.diagnostics["count"] = xr.DataArray(
                bur.get_count(), dims=("y", "x"), attrs={"area": self.base_area})
        return (inv_lon, inv_lat)


class ParallaxCorrectionModifier(ModifierBase):
    """Modifier for parallax correction.

    Apply parallax correction as a modifier.  Uses the
    :class:`ParallaxCorrection` class, which in turn uses the
    :func:`get_parallax_corrected_lonlats` function.  See the documentation there for
    details on the behaviour.

    To use this, add to ``composites/visir.yaml`` within ``SATPY_CONFIG_PATH``
    something like::

        sensor_name: visir

        modifiers:
          parallax_corrected:
            modifier: !!python/name:satpy.modifiers.parallax.ParallaxCorrectionModifier
            prerequisites:
              - "ctth_alti"
            dataset_radius_of_influence: 50000

        composites:

          parallax_corrected_VIS006:
            compositor: !!python/name:satpy.composites.core.SingleBandCompositor
            prerequisites:
              - name: VIS006
                modifiers: [parallax_corrected]

    Here, ``ctth_alti`` is CTH provided by the ``nwcsaf-geo`` reader, so to use it
    one would have to pass both on scene creation::

        sc = Scene({"seviri_l1b_hrit": files_l1b, "nwcsaf-geo": files_l2})
        sc.load(["parallax_corrected_VIS006"])

    The modifier takes optional global parameters, all of which are optional.
    They affect various steps in the algorithm.  Setting them may impact
    performance:

    cth_resampler
        Resampler to use when resampling (cloud top) height to the base area.
        Defaults to "nearest".
    cth_radius_of_influence
        Radius of influence to use when resampling the (cloud top) height to
        the base area.  Defaults to 50000.
    lonlat_chunks
        Chunk size to use when obtaining longitudes and latitudes from the area
        definition.  Defaults to 1024.  If you set this to None, then parallax
        correction will involve premature calculation.  Changing this may or
        may not make parallax correction slower or faster.
    dataset_radius_of_influence
        Radius of influence to use when resampling the dataset onto the
        swathdefinition describing the parallax-corrected area.  Defaults to
        50000.  This always uses nearest neighbour resampling.

    Alternately, you can use the lower-level API directly with the
    :class:`ParallaxCorrection` class, which may be more efficient if multiple
    datasets need to be corrected.  RGB Composites cannot be modified in this way
    (i.e. you can't replace "VIS006" by "natural_color").  To get a parallax
    corrected RGB composite, create a new composite where each input has the
    modifier applied.  The parallax calculation should only occur once, because
    calculations are happening via dask and dask should reuse the calculation.
    """

    def __call__(self, projectables, optional_datasets=None, **info):
        """Apply parallax correction.

        The argument ``projectables`` needs to contain the dataset to be
        projected and the height to use for the correction.
        """
        (to_be_corrected, cth) = projectables
        base_area = to_be_corrected.attrs["area"]
        corrector = self._get_corrector(base_area)
        plax_corr_area = corrector(
                cth,
                cth_resampler=self.attrs.get("cth_resampler", "nearest"),
                cth_radius_of_influence=self.attrs.get("cth_radius_of_influence", 50_000),
                lonlat_chunks=self.attrs.get("lonlat_chunks", 1024),
                )
        res = resample_dataset(
                to_be_corrected, plax_corr_area,
                radius_of_influence=self.attrs.get("dataset_radius_of_influence", 50_000),
                fill_value=np.nan)
        res.attrs["area"] = to_be_corrected.attrs["area"]
        self.apply_modifier_info(to_be_corrected, res)

        return res

    def _get_corrector(self, base_area):
        # only pass on those attributes that are arguments by
        # ParallaxCorrection.__init__
        sig = inspect.signature(ParallaxCorrection.__init__)
        kwargs = {}
        for k in sig.parameters.keys() & self.attrs.keys():
            kwargs[k] = self.attrs[k]
        corrector = ParallaxCorrection(base_area, **kwargs)
        return corrector


def _get_satpos_from_cth(cth_dataset):
    """Obtain satellite position from CTH dataset.

    From a CTH dataset, obtain the satellite position lon, lat, altitude [m],
    either directly from orbital parameters, or, when missing, from the
    platform name using pyorbital and skyfield.
    """
    (sat_lon, sat_lat, sat_alt) = get_satpos(
            cth_dataset, use_tle=True)
    return (sat_lon, sat_lat, sat_alt)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Benchmark AHI HSD operations.."""

from __future__ import annotations

import os

from pyspectral.rayleigh import check_and_download as download_luts
from pyspectral.rsr_reader import check_and_download as download_rsr

from benchmarks.utils import GeoBenchmarks, get_filenames


class HimawariHSD(GeoBenchmarks):
    """Benchmark Himawari HSD reading."""

    timeout = 600
    data_files: list[str] = []
    subdir = os.path.join("ahi_hsd", "20210417_0500_typhoon_surigae")
    reader = "ahi_hsd"

    def setup_cache(self):
        """Fetch the data files."""
        try:
            from satpy.demo import download_typhoon_surigae_ahi
            download_typhoon_surigae_ahi(channels=[1, 2, 3, 4], segments=[4])
        except ImportError:
            assert len(get_filenames(self.subdir)) == 4  # nosec
        download_rsr()
        download_luts(aerosol_types=["rayleigh_only"])

    def setup(self):
        """Set up the benchmarks."""
        import dask.config
        self.data_files = get_filenames(self.subdir)
        dask.config.set({"array.chunk-size": "32MiB"})

    def time_load_one_channel(self):
        """Time the loading of one channel."""
        self.compute_channel("B01")

    def peakmem_load_one_channel(self):
        """Check peak memory usage of loading one channel."""
        self.compute_channel("B01")

    def time_load_true_color(self):
        """Time the loading of the generation of true_color."""
        self.compute_composite("true_color")

    def peakmem_load_true_color(self):
        """Check peak memory usage of the generation of true_color."""
        self.compute_composite("true_color")

    def time_save_true_color_nocorr_to_geotiff(self):
        """Time the generation and saving of true_color_nocorr."""
        self.save_composite_as_geotiff("true_color_nocorr")

    def peakmem_save_true_color_to_geotiff(self):
        """Check peak memory usage of the generation and saving of true_color_nocorr."""
        self.save_composite_as_geotiff("true_color_nocorr")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2012 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Reads a format from an xml file to create dtypes and scaling factor arrays."""

from __future__ import annotations

import numpy as np
from defusedxml.ElementTree import parse

VARIABLES: dict[str, str] = {}

TYPEC = {"boolean": ">i1",
         "integer2": ">i2",
         "integer4": ">i4",
         "uinteger2": ">u2",
         "uinteger4": ">u4", }


def process_delimiter(elt, text=False):
    """Process a 'delimiter' tag."""
    del elt, text


def process_field(elt, text=False):
    """Process a 'field' tag."""
    # NOTE: if there is a variable defined in this field and it is different
    # from the default, we could change the value and restart.

    scale = np.uint8(1)
    if elt.get("type") == "bitfield" and not text:
        current_type = ">u" + str(int(elt.get("length")) // 8)
        scale = np.dtype(current_type).type(1)
    elif (elt.get("length") is not None):
        if text:
            add = 33
        else:
            add = 0
        current_type = "S" + str(int(elt.get("length")) + add)
    else:
        current_type = TYPEC[elt.get("type")]
        try:
            scale = (10 /
                     float(elt.get("scaling-factor", "10").replace("^", "e")))
        except ValueError:
            scale = (10 / np.array(
                elt.get("scaling-factor").replace("^", "e").split(","),
                dtype=np.float64))

    return ((elt.get("name"), current_type, scale))


def process_array(elt, text=False):
    """Process an 'array' tag."""
    del text
    chld = list(elt)
    if len(chld) > 1:
        raise ValueError()
    chld = chld[0]
    try:
        name, current_type, scale = CASES[chld.tag](chld)
        size = None
    except ValueError:
        name, current_type, size, scale = CASES[chld.tag](chld)
    del name
    myname = elt.get("name") or elt.get("label")
    if elt.get("length").startswith("$"):
        length = int(VARIABLES[elt.get("length")[1:]])
    else:
        length = int(elt.get("length"))
    if size is not None:
        return (myname, current_type, (length, ) + size, scale)
    else:
        return (myname, current_type, (length, ), scale)


CASES = {"delimiter": process_delimiter,
         "field": process_field,
         "array": process_array, }


def to_dtype(val):
    """Parse *val* to return a dtype."""
    return np.dtype([i[:-1] for i in val])


def to_scaled_dtype(val):
    """Parse *val* to return a dtype."""
    res = []
    for i in val:
        if i[1].startswith("S"):
            res.append((i[0], i[1]) + i[2:-1])
        else:
            try:
                res.append((i[0], i[-1].dtype) + i[2:-1])
            except AttributeError:
                res.append((i[0], type(i[-1])) + i[2:-1])

    return np.dtype(res)


def to_scales(val):
    """Parse *val* to return an array of scale factors."""
    res = []
    for i in val:
        if len(i) == 3:
            res.append((i[0], type(i[2])))
        else:
            try:
                res.append((i[0], i[3].dtype, i[2]))
            except AttributeError:
                res.append((i[0], type(i[3]), i[2]))

    dtype = np.dtype(res)

    scales = np.zeros((1, ), dtype=dtype)

    for i in val:
        try:
            scales[i[0]] = i[-1]
        except ValueError:
            scales[i[0]] = np.repeat(np.array(i[-1]), i[2][1]).reshape(i[2])

    return scales


def parse_format(xml_file):
    """Parse the xml file to create types, scaling factor types, and scales."""
    tree = parse(xml_file)
    for param in tree.find("parameters"):
        VARIABLES[param.get("name")] = param.get("value")

    types_scales = {}

    for prod in tree.find("product"):
        text = (prod.tag in ["mphr", "sphr"])
        res = []
        for i in prod:
            lres = CASES[i.tag](i, text)
            if lres is not None:
                res.append(lres)
        types_scales[(prod.tag, int(prod.get("subclass")))] = res

    types = {}
    stypes = {}
    scales = {}

    for key, val in types_scales.items():
        types[key] = to_dtype(val)
        stypes[key] = to_scaled_dtype(val)
        scales[key] = to_scales(val)

    return types, stypes, scales


def _apply_scales(array, scales, dtype):
    """Apply scales to the array."""
    new_array = np.empty(array.shape, dtype)
    for i in array.dtype.names:
        try:
            new_array[i] = array[i] * scales[i]
        except TypeError:
            if np.all(scales[i] == 1):
                new_array[i] = array[i]
            else:
                raise
    return new_array


class XMLFormat(object):
    """XMLFormat object."""

    def __init__(self, filename):
        """Init the format reader."""
        self.types, self.stypes, self.scales = parse_format(filename)

        self.translator = {}

        for key, val in self.types.items():
            self.translator[val] = (self.scales[key], self.stypes[key])

    def dtype(self, key):
        """Get the dtype for the format object."""
        return self.types[key]

    def apply_scales(self, array):
        """Apply scales to *array*."""
        return _apply_scales(array, *self.translator[array.dtype])


if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Advanced Geostationary Radiation Imager reader for the Level_1 HDF format.

The files read by this reader are described in the official Real Time Data Service:

    http://fy4.nsmc.org.cn/data/en/data/realtime.html

"""

import logging

from satpy.readers.core.fy4 import FY4Base

logger = logging.getLogger(__name__)


class HDF_AGRI_L1(FY4Base):
    """AGRI l1 file handler."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init filehandler."""
        super(HDF_AGRI_L1, self).__init__(filename, filename_info, filetype_info)
        self.sensor = "AGRI"

    def get_dataset(self, dataset_id, ds_info):
        """Load a dataset."""
        ds_name = dataset_id["name"]
        logger.debug("Reading in get_dataset %s.", ds_name)
        file_key = ds_info.get("file_key", ds_name)
        if self.PLATFORM_ID == "FY-4B":
            if self.CHANS_ID in file_key:
                file_key = f"Data/{file_key}"
            elif self.SUN_ID in file_key or self.SAT_ID in file_key:
                file_key = f"Navigation/{file_key}"
        data = self.get(file_key)
        if data.ndim >= 2:
            data = data.rename({data.dims[-2]: "y", data.dims[-1]: "x"})
        data = self.calibrate(data, ds_info, ds_name, file_key)

        self.adjust_attrs(data, ds_info)

        return data

    def adjust_attrs(self, data, ds_info):
        """Adjust the attrs of the data."""
        satname = self.PLATFORM_NAMES.get(self["/attr/Satellite Name"], self["/attr/Satellite Name"])
        data.attrs.update({"platform_name": satname,
                           "sensor": self["/attr/Sensor Identification Code"].lower(),
                           "orbital_parameters": {
                               "satellite_nominal_latitude": self["/attr/NOMCenterLat"].item(),
                               "satellite_nominal_longitude": self["/attr/NOMCenterLon"].item(),
                               "satellite_nominal_altitude": self["/attr/NOMSatHeight"].item()}})
        data.attrs.update(ds_info)
        # remove attributes that could be confusing later
        data.attrs.pop("FillValue", None)
        data.attrs.pop("Intercept", None)
        data.attrs.pop("Slope", None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""MODIS L1b and L2 test fixtures."""

from __future__ import annotations

import datetime as dt
from typing import Literal, Optional

import numpy as np
import pytest
from pyhdf.SD import SD, SDC

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmpdir_factory

# Level 1 Fixtures

AVAILABLE_1KM_VIS_PRODUCT_NAMES = [str(x) for x in range(8, 13)]
AVAILABLE_1KM_VIS_PRODUCT_NAMES += ["13lo", "13hi", "14lo", "14hi"]
AVAILABLE_1KM_VIS_PRODUCT_NAMES += [str(x) for x in range(15, 20)]
AVAILABLE_1KM_IR_PRODUCT_NAMES = [str(x) for x in range(20, 37)]
AVAILABLE_1KM_PRODUCT_NAMES = AVAILABLE_1KM_VIS_PRODUCT_NAMES + AVAILABLE_1KM_IR_PRODUCT_NAMES
AVAILABLE_HKM_PRODUCT_NAMES = [str(x) for x in range(3, 8)]
AVAILABLE_QKM_PRODUCT_NAMES = ["1", "2"]
SCAN_LEN_5KM = 6  # 3 scans of 5km data
SCAN_WIDTH_5KM = 270
SCALE_FACTOR = 0.5
ADD_OFFSET = -0.5
RES_TO_REPEAT_FACTOR = {
    250: 20,
    500: 10,
    1000: 5,
    5000: 1,
}


def _shape_for_resolution(resolution: int) -> tuple[int, int]:
    assert resolution in RES_TO_REPEAT_FACTOR
    factor = RES_TO_REPEAT_FACTOR[resolution]
    if factor == 1:
        return SCAN_LEN_5KM, SCAN_WIDTH_5KM

    factor_1km = RES_TO_REPEAT_FACTOR[1000]
    shape_1km = (factor_1km * SCAN_LEN_5KM, factor_1km * SCAN_WIDTH_5KM + 4)
    factor //= 5
    return factor * shape_1km[0], factor * shape_1km[1]


def _generate_lonlat_data(resolution: int) -> tuple[np.ndarray, np.ndarray]:
    shape = _shape_for_resolution(resolution)
    lat = np.repeat(np.linspace(35., 45., shape[0])[:, None], shape[1], 1)
    lat *= np.linspace(0.9, 1.1, shape[1])
    lon = np.repeat(np.linspace(-45., -35., shape[1])[None, :], shape[0], 0)
    lon *= np.linspace(0.9, 1.1, shape[0])[:, None]
    return lon.astype(np.float32), lat.astype(np.float32)


def _generate_angle_data(resolution: int) -> np.ndarray:
    shape = _shape_for_resolution(resolution)
    data = np.repeat(abs(np.linspace(-65.2, 65.4, shape[1]))[None, :], shape[0], 0)
    return (data * 100).astype(np.int16)


def _generate_visible_data(resolution: int, num_bands: int, dtype=np.uint16) -> np.ndarray:
    shape = _shape_for_resolution(resolution)
    data = np.ones((num_bands, shape[0], shape[1]), dtype=dtype)

    # add fill value to every band
    data[:, -1, -1] = 65535

    # add band 2 saturation and can't aggregate fill values
    data[1, -1, -2] = 65533
    data[1, -1, -3] = 65528
    return data


def _generate_visible_uncertainty_data(shape: tuple) -> np.ndarray:
    uncertainty = np.zeros(shape, dtype=np.uint8)
    uncertainty[:, -1, -1] = 15  # fill value
    uncertainty[:, -1, -2] = 15  # saturated
    uncertainty[:, -1, -3] = 15  # can't aggregate
    return uncertainty


def _get_lonlat_variable_info(resolution: int) -> dict:
    lon_5km, lat_5km = _generate_lonlat_data(resolution)
    return {
        "Latitude": {"data": lat_5km,
                     "type": SDC.FLOAT32,
                     "fill_value": -999,
                     "attrs": {"dim_labels": ["Cell_Along_Swath_5km:mod35", "Cell_Across_Swath_5km:mod35"]}},
        "Longitude": {"data": lon_5km,
                      "type": SDC.FLOAT32,
                      "fill_value": -999,
                      "attrs": {"dim_labels": ["Cell_Along_Swath_5km:mod35", "Cell_Across_Swath_5km:mod35"]}},
    }


def _get_angles_variable_info(resolution: int) -> dict:
    angle_data = _generate_angle_data(resolution)
    dim_factor = RES_TO_REPEAT_FACTOR[resolution] * 2
    angle_info = {
        "data": angle_data,
        "type": SDC.INT16,
        "fill_value": -32767,
        "attrs": {
            "dim_labels": [
                f"{dim_factor}*nscans:MODIS_SWATH_Type_L1B",
                "1KM_geo_dim:MODIS_SWATH_Type_L1B"],
            "scale_factor": 0.01,
            "add_offset": -0.01,
        },
    }
    angles_info = {}
    for var_name in ("SensorAzimuth", "SensorZenith", "SolarAzimuth", "SolarZenith"):
        angles_info[var_name] = angle_info
    return angles_info


def _get_visible_variable_info(var_name: str, resolution: int, bands: list[str]):
    num_bands = len(bands)
    data = _generate_visible_data(resolution, len(bands))
    uncertainty = _generate_visible_uncertainty_data(data.shape)
    dim_factor = RES_TO_REPEAT_FACTOR[resolution] * 2
    band_dim_name = f"Band_{resolution}_{num_bands}_RefSB:MODIS_SWATH_Type_L1B"
    row_dim_name = f"{dim_factor}*nscans:MODIS_SWATH_Type_L1B"
    col_dim_name = "Max_EV_frames:MODIS_SWATH_Type_L1B"
    return {
        var_name: {
            "data": data,
            "type": SDC.UINT16,
            "fill_value": 0,
            "attrs": {
                # dim_labels are just unique dimension names, may not match exactly with real world files
                "dim_labels": [band_dim_name,
                               row_dim_name,
                               col_dim_name],
                "valid_range": (0, 32767),
                "reflectance_scales": (2.0,) * num_bands,
                "reflectance_offsets": (-0.5,) * num_bands,
                "band_names": ",".join(bands),
            },
        },
        var_name + "_Uncert_Indexes": {
            "data": uncertainty,
            "type": SDC.UINT8,
            "fill_value": 255,
            "attrs": {
                "dim_labels": [band_dim_name,
                               row_dim_name,
                               col_dim_name],
            },
        },
    }


def _get_emissive_variable_info(var_name: str, resolution: int, bands: list[str]):
    num_bands = len(bands)
    data = _generate_visible_data(resolution, len(bands))
    dim_factor = RES_TO_REPEAT_FACTOR[resolution] * 2
    band_dim_name = f"Band_{resolution}_{num_bands}_Emissive:MODIS_SWATH_Type_L1B"
    row_dim_name = f"{dim_factor}*nscans:MODIS_SWATH_Type_L1B"
    col_dim_name = "Max_EV_frames:MODIS_SWATH_Type_L1B"
    return {
        var_name: {
            "data": data,
            "type": SDC.UINT16,
            "fill_value": 0,
            "attrs": {
                "dim_labels": [band_dim_name,
                               row_dim_name,
                               col_dim_name],
                "valid_range": (0, 32767),
                "band_names": ",".join(bands),
            },
        },
        var_name + "_Uncert_Indexes": {
            "data": np.zeros(data.shape, dtype=np.uint8),
            "type": SDC.UINT8,
            "fill_value": 255,
            "attrs": {
                "dim_labels": [band_dim_name,
                               row_dim_name,
                               col_dim_name],
            },
        },
    }


def _get_l1b_geo_variable_info(filename: str,
                               geo_resolution: int,
                               include_angles: bool = True
                               ) -> dict:
    variables_info = {}
    variables_info.update(_get_lonlat_variable_info(geo_resolution))
    if include_angles:
        variables_info.update(_get_angles_variable_info(geo_resolution))
    return variables_info

def _get_l3_land_cover_info() -> dict:

    lc_data = np.zeros((2400, 2400), dtype=np.uint8)

    variables_info = \
    {
        "LC_Type1": {"data": lc_data,
                     "type": SDC.UINT8,
                     "fill_value": 255,
                     "attrs": {
                         "dim_labels": ["YDim:MCD12Q1", "XDim:MCD12Q1"],
                     },
                     },
        "LC_Type2": {"data": lc_data,
                     "type": SDC.UINT8,
                     "fill_value": 255,
                     "attrs": {
                         "dim_labels": ["YDim:MCD12Q1", "XDim:MCD12Q1"],
                     },
                     },
    }

    return variables_info


def generate_nasa_l1b_filename(prefix):
    """Generate a filename that follows NASA MODIS L1b convention."""
    now = dt.datetime.now()
    return f"{prefix}_A{now:%y%j_%H%M%S}_{now:%Y%j%H%M%S}.hdf"


def generate_imapp_filename(suffix):
    """Generate a filename that follows IMAPP MODIS L1b convention."""
    now = dt.datetime.now()
    return f"t1.{now:%y%j.%H%M}.{suffix}.hdf"


def create_hdfeos_test_file(filename: str,
                            variable_infos: dict,
                            struct_meta: Optional[str] = None,
                            core_meta: Optional[str] = None,
                            archive_meta: Optional[str] = None,
                            ) -> None:
    """Create a fake MODIS L1b HDF4 file with headers.

    Args:
        filename: Full path of filename to be created.
        variable_infos: Dictionary mapping HDF4 variable names to dictionary
            of variable information (see ``_add_variable_to_file``).
        struct_meta: Contents of the 'StructMetadata.0' header.
        core_meta: Contents of the 'CoreMetadata.0' header.
        archive_meta:Contents of the 'ArchiveMetadata.0' header.

    """
    h = SD(filename, SDC.WRITE | SDC.CREATE)

    if struct_meta:
        setattr(h, "StructMetadata.0", struct_meta)
    if core_meta:
        setattr(h, "CoreMetadata.0", core_meta)
    if archive_meta:
        setattr(h, "ArchiveMetadata.0", archive_meta)

    for var_name, var_info in variable_infos.items():
        _add_variable_to_file(h, var_name, var_info)

    h.end()


def _add_variable_to_file(h, var_name, var_info):
    v = h.create(var_name, var_info["type"], var_info["data"].shape)
    v[:] = var_info["data"]
    dim_count = 0
    for dimension_name in var_info["attrs"]["dim_labels"]:
        v.dim(dim_count).setname(dimension_name)
        dim_count += 1
    v.setfillvalue(var_info["fill_value"])
    v.scale_factor = var_info["attrs"].get("scale_factor", SCALE_FACTOR)
    v.add_offset = var_info["attrs"].get("add_offset", ADD_OFFSET)
    for attr_key, attr_val in var_info["attrs"].items():
        if attr_key == "dim_labels":
            continue
        setattr(v, attr_key, attr_val)


def _create_core_metadata(file_shortname: str) -> str:
    beginning_date = dt.datetime.now()
    ending_date = beginning_date + dt.timedelta(minutes=5)
    core_metadata_header = "GROUP = INVENTORYMETADATA\nGROUPTYPE = MASTERGROUP\n\n" \
                           'GROUP = RANGEDATETIME\n\nOBJECT = RANGEBEGINNINGDATE\nNUM_VAL = 1\nVALUE = "{}"\n' \
                           "END_OBJECT = RANGEBEGINNINGDATE\n\nOBJECT = RANGEBEGINNINGTIME\n" \
                           'NUM_VAL = 1\nVALUE = "{}"\n' \
                           "END_OBJECT = RANGEBEGINNINGTIME\n\nOBJECT = RANGEENDINGDATE\n" \
                           'NUM_VAL = 1\nVALUE = "{}"\n' \
                           'END_OBJECT = RANGEENDINGDATE\n\nOBJECT = RANGEENDINGTIME\nNUM_VAL = 1\nVALUE = "{}"\n' \
                           "END_OBJECT = RANGEENDINGTIME\nEND_GROUP = RANGEDATETIME"
    core_metadata_header = core_metadata_header.format(
        beginning_date.strftime("%Y-%m-%d"),
        beginning_date.strftime("%H:%M:%S.%f"),
        ending_date.strftime("%Y-%m-%d"),
        ending_date.strftime("%H:%M:%S.%f")
    )
    inst_metadata = "GROUP = ASSOCIATEDPLATFORMINSTRUMENTSENSOR\n\n" \
                    'OBJECT = ASSOCIATEDPLATFORMINSTRUMENTSENSORCONTAINER\nCLASS = "1"\n\n' \
                    'OBJECT = ASSOCIATEDSENSORSHORTNAME\nCLASS = "1"\nNUM_VAL = 1\n' \
                    'VALUE = "MODIS"\nEND_OBJECT = ASSOCIATEDSENSORSHORTNAME\n\n' \
                    'OBJECT = ASSOCIATEDPLATFORMSHORTNAME\nCLASS = "1"\nNUM_VAL = 1\n' \
                    'VALUE = "Terra"\nEND_OBJECT = ASSOCIATEDPLATFORMSHORTNAME\n\n' \
                    'OBJECT = ASSOCIATEDINSTRUMENTSHORTNAME\nCLASS = "1"\nNUM_VAL = 1\n' \
                    'VALUE = "MODIS"\nEND_OBJECT = ASSOCIATEDINSTRUMENTSHORTNAME\n\n' \
                    "END_OBJECT = ASSOCIATEDPLATFORMINSTRUMENTSENSORCONTAINER\n\n" \
                    "END_GROUP              = ASSOCIATEDPLATFORMINSTRUMENTSENSOR\n\n"
    collection_metadata = "GROUP = COLLECTIONDESCRIPTIONCLASS\n\nOBJECT = SHORTNAME\nNUM_VAL = 1\n" \
                          f"VALUE = {file_shortname!r}\nEND_OBJECT = SHORTNAME\n\n" \
                          "OBJECT = VERSIONID\nNUM_VAL = 1\nVALUE = 6\nEND_OBJECT = VERSIONID\n\n" \
                          "END_GROUP = COLLECTIONDESCRIPTIONCLASS\n\n"
    core_metadata_header += "\n\n" + inst_metadata + collection_metadata
    return core_metadata_header


def _create_struct_metadata(geo_resolution: int) -> str:
    geo_dim_factor = RES_TO_REPEAT_FACTOR[geo_resolution] * 2
    struct_metadata_header = "GROUP=SwathStructure\n" \
                             "GROUP=SWATH_1\n" \
                             "GROUP=DimensionMap\n" \
                             "OBJECT=DimensionMap_2\n" \
                             f'GeoDimension="{geo_dim_factor}*nscans"\n' \
                             "END_OBJECT=DimensionMap_2\n" \
                             "END_GROUP=DimensionMap\n" \
                             "END_GROUP=SWATH_1\n" \
                             "END_GROUP=SwathStructure\nEND"
    return struct_metadata_header


def _create_struct_metadata_cmg(ftype: str) -> str:
    # Case of a MOD09 file
    if ftype == "MOD09":
        gridline = 'GridName="MOD09CMG"\n'
        upleft = "UpperLeftPointMtrs=(-180000000.000000,90000000.000000)\n"
        upright = "LowerRightMtrs=(180000000.000000,-90000000.000000)\n"
        XDim=7200
        YDim=3600
    # Case of a MCD12Q1 file
    elif ftype == "MCD12Q1":
        gridline = 'GridName="MCD12Q1"\n'
        upleft = "UpperLeftPointMtrs=(-8895604.157333,-1111950.519667)\n"
        upright = "LowerRightMtrs=(-7783653.637667,-2223901.039333)\n"
        XDim=2400
        YDim=2400
    # Case of a MCD43 file
    else:
        gridline = 'GridName="MCD_CMG_BRDF_0.05Deg"\n'
        upleft = "UpperLeftPointMtrs=(-180.000000,90.000000)\n"
        upright = "LowerRightMtrs=(180.000000,-90.000000)\n"
        XDim=7200
        YDim=3600

    struct_metadata_header = ("GROUP=SwathStructure\n"
                              "END_GROUP=SwathStructure\n"
                              "GROUP=GridStructure\n"
                              "GROUP=GRID_1\n"
                              f"{gridline}\n"
                              f"XDim={XDim}\n"
                              f"YDim={YDim}\n"
                              f"{upleft}\n"
                              f"{upright}\n"
                              "END_GROUP=GRID_1\n"
                              "END_GROUP=GridStructure\nEND")
    return struct_metadata_header


def _create_header_metadata() -> str:
    archive_metadata_header = ("GROUP = ARCHIVEDMETADATA\n"
                               'TEST_URL = "http://modis.gsfc.nasa.gov/?some_val=100"\n'
                               "END_GROUP = ARCHIVEDMETADATA\nEND")
    return archive_metadata_header


@pytest.fixture(scope="session")
def modis_l1b_nasa_mod021km_file(tmpdir_factory) -> list[str]:
    """Create a single MOD021KM file following standard NASA file scheme."""
    filename = generate_nasa_l1b_filename("MOD021km")
    full_path = str(tmpdir_factory.mktemp("modis_l1b").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 5000, include_angles=True)
    variable_infos.update(_get_visible_variable_info("EV_1KM_RefSB", 1000, AVAILABLE_1KM_VIS_PRODUCT_NAMES))
    variable_infos.update(_get_visible_variable_info("EV_500_Aggr1km_RefSB", 1000, AVAILABLE_HKM_PRODUCT_NAMES))
    variable_infos.update(_get_visible_variable_info("EV_250_Aggr1km_RefSB", 1000, AVAILABLE_QKM_PRODUCT_NAMES))
    variable_infos.update(_get_emissive_variable_info("EV_1KM_Emissive", 1000, AVAILABLE_1KM_IR_PRODUCT_NAMES))
    create_hdfeos_test_file(full_path,
                            variable_infos,
                            _create_struct_metadata(5000),
                            _create_core_metadata("MOD021KM"),
                            _create_header_metadata())
    return [full_path]


@pytest.fixture(scope="session")
def modis_l1b_imapp_1000m_file(tmpdir_factory) -> list[str]:
    """Create a single MOD021KM file following IMAPP file scheme."""
    filename = generate_imapp_filename("1000m")
    full_path = str(tmpdir_factory.mktemp("modis_l1b").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 5000, include_angles=True)
    variable_infos.update(_get_visible_variable_info("EV_1KM_RefSB", 1000, AVAILABLE_1KM_VIS_PRODUCT_NAMES))
    variable_infos.update(_get_visible_variable_info("EV_500_Aggr1km_RefSB", 1000, AVAILABLE_HKM_PRODUCT_NAMES))
    variable_infos.update(_get_visible_variable_info("EV_250_Aggr1km_RefSB", 1000, AVAILABLE_QKM_PRODUCT_NAMES))
    variable_infos.update(_get_emissive_variable_info("EV_1KM_Emissive", 1000, AVAILABLE_1KM_IR_PRODUCT_NAMES))
    create_hdfeos_test_file(full_path,
                            variable_infos,
                            _create_struct_metadata(5000),
                            _create_core_metadata("MOD021KM"),
                            _create_header_metadata())
    return [full_path]


@pytest.fixture(scope="session")
def modis_l1b_nasa_mod02hkm_file(tmpdir_factory) -> list[str]:
    """Create a single MOD02HKM file following standard NASA file scheme."""
    filename = generate_nasa_l1b_filename("MOD02Hkm")
    full_path = str(tmpdir_factory.mktemp("modis_l1b").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 1000, include_angles=False)
    variable_infos.update(_get_visible_variable_info("EV_500_RefSB", 250, AVAILABLE_QKM_PRODUCT_NAMES))
    create_hdfeos_test_file(full_path,
                            variable_infos,
                            _create_struct_metadata(1000),
                            _create_core_metadata("MOD02HKM"),
                            _create_header_metadata())
    return [full_path]


@pytest.fixture
def modis_l1b_nasa_mod02qkm_file(tmpdir_factory) -> list[str]:
    """Create a single MOD02QKM file following standard NASA file scheme."""
    filename = generate_nasa_l1b_filename("MOD02Qkm")
    full_path = str(tmpdir_factory.mktemp("modis_l1b").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 1000, include_angles=False)
    variable_infos.update(_get_visible_variable_info("EV_250_RefSB", 250, AVAILABLE_QKM_PRODUCT_NAMES))
    create_hdfeos_test_file(full_path,
                            variable_infos,
                            _create_struct_metadata(1000),
                            _create_core_metadata("MOD02QKM"),
                            _create_header_metadata())
    return [full_path]


@pytest.fixture(scope="session")
def modis_l1b_nasa_mod03_file(tmpdir_factory) -> list[str]:
    """Create a single MOD03 file following standard NASA file scheme."""
    filename = generate_nasa_l1b_filename("MOD03")
    full_path = str(tmpdir_factory.mktemp("modis_l1b").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 1000, include_angles=True)
    create_hdfeos_test_file(full_path,
                            variable_infos,
                            _create_struct_metadata(1000),
                            _create_core_metadata("MOD03"),
                            _create_header_metadata())
    return [full_path]


@pytest.fixture(scope="session")
def modis_l1b_imapp_geo_file(tmpdir_factory) -> list[str]:
    """Create a single geo file following standard IMAPP file scheme."""
    filename = generate_imapp_filename("geo")
    full_path = str(tmpdir_factory.mktemp("modis_l1b").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 1000, include_angles=True)
    create_hdfeos_test_file(full_path,
                            variable_infos,
                            _create_struct_metadata(1000),
                            _create_core_metadata("MOD03"),
                            _create_header_metadata())
    return [full_path]


@pytest.fixture(scope="session")
def modis_l1b_nasa_1km_mod03_files(modis_l1b_nasa_mod021km_file, modis_l1b_nasa_mod03_file) -> list[str]:
    """Create input files including the 1KM and MOD03 files."""
    return modis_l1b_nasa_mod021km_file + modis_l1b_nasa_mod03_file


# Level 2 Fixtures


def _get_basic_variable_info(var_name: str, resolution: int, dim_size: Literal[2, 3]=2) -> dict:
    shape = _shape_for_resolution(resolution)

    row_dim_name = f"Cell_Along_Swath_{resolution}m:modl2"
    col_dim_name = f"Cell_Across_Swath_{resolution}m:modl2"

    if dim_size == 3:
        data = np.ones((1, shape[0], shape[1]), dtype=np.uint16)
        dim_labels = ["channel", row_dim_name, col_dim_name]
    elif dim_size == 2:
        data = np.ones((shape[0], shape[1]), dtype=np.uint16)
        dim_labels = [row_dim_name, col_dim_name]


    return {
        var_name: {
            "data": data,
            "type": SDC.UINT16,
            "fill_value": 0,
            "attrs": {
                # dim_labels are just unique dimension names, may not match exactly with real world files
                "dim_labels": dim_labels,
                "valid_range": (0, 32767),
                "scale_factor": 2.0,
                "add_offset": -1.0,
            },
        },
    }


def _get_cloud_mask_variable_info(var_name: str, resolution: int) -> dict:
    num_bytes = 6
    shape = _shape_for_resolution(resolution)
    data = np.zeros((num_bytes, shape[0], shape[1]), dtype=np.int8)
    byte_dim_name = "Byte_Segment:mod35"
    row_dim_name = "Cell_Along_Swath_1km:mod35"
    col_dim_name = "Cell_Across_Swath_1km:mod35"
    return {
        var_name: {
            "data": data,
            "type": SDC.INT8,
            "fill_value": 0,
            "attrs": {
                # dim_labels are just unique dimension names, may not match exactly with real world files
                "dim_labels": [byte_dim_name,
                               row_dim_name,
                               col_dim_name],
                "valid_range": (0, -1),
                "scale_factor": 1.,
                "add_offset": 0.,
            },
        },
        "Quality_Assurance": {
            "data": np.ones((shape[0], shape[1], 10), dtype=np.int8),
            "type": SDC.INT8,
            "fill_value": 0,
            "attrs": {
                # dim_labels are just unique dimension names, may not match exactly with real world files
                "dim_labels": [row_dim_name,
                               col_dim_name,
                               "Quality_Dimension:mod35"],
                "valid_range": (0, -1),
                "scale_factor": 2.,
                "add_offset": -0.5,
            },
        },
    }


def _get_l3_refl_variable_info(var_name: str) -> dict:
    shape = (3600, 7200)
    data = np.zeros((shape[0], shape[1]), dtype=np.int16)
    row_dim_name = "XDim"
    col_dim_name = "YDim"
    return {
        var_name: {
            "data": data,
            "type": SDC.INT16,
            "fill_value": -28672,
            "attrs": {
                # dim_labels are just unique dimension names, may not match exactly with real world files
                "dim_labels": [row_dim_name,
                               col_dim_name],
                "valid_range": (-100, 16000),
                "scale_factor": 1e-4,
                "add_offset": 0.,
            },
        },
    }


def _get_mask_byte1_variable_info() -> dict:
    shape = _shape_for_resolution(1000)
    data = np.zeros((shape[0], shape[1]), dtype=np.uint16)
    row_dim_name = "Cell_Along_Swath_1km:mod35"
    col_dim_name = "Cell_Across_Swath_1km:mod35"
    return {
        "MODIS_Cloud_Mask": {
            "data": data,
            "type": SDC.UINT16,
            "fill_value": 9999,
            "attrs": {
                # dim_labels are just unique dimension names, may not match exactly with real world files
                "dim_labels": [row_dim_name,
                               col_dim_name],
                "valid_range": (0, 4),
                "scale_factor": 2,
                "add_offset": -1,
            },

        },
        "MODIS_Simple_LandSea_Mask": {
            "data": data,
            "type": SDC.UINT16,
            "fill_value": 9999,
            "attrs": {
                # dim_labels are just unique dimension names, may not match exactly with real world files
                "dim_labels": [row_dim_name,
                               col_dim_name],
                "valid_range": (0, 4),
                "scale_factor": 2,
                "add_offset": -1,
            },
        },
        "MODIS_Snow_Ice_Flag": {
            "data": data,
            "type": SDC.UINT16,
            "fill_value": 9999,
            "attrs": {
                # dim_labels are just unique dimension names, may not match exactly with real world files
                "dim_labels": [row_dim_name,
                               col_dim_name],
                "valid_range": (0, 2),
                "scale_factor": 2,
                "add_offset": -1,
            },
        },
    }


def generate_nasa_l2_filename(prefix: str) -> str:
    """Generate a file name that follows MODIS 35 L2 convention in a temporary directory."""
    now = dt.datetime.now()
    return f"{prefix}_L2.A{now:%Y%j.%H%M}.061.{now:%Y%j%H%M%S}.hdf"

def generate_nasa_l3_tile_filename(prefix: str) -> str:
    """Generate a file name that follows MODIS sinusoidal grid tile pattern."""
    now = dt.datetime.now()
    return f"{prefix}.A{now:%Y}001.h34v07.061.{now:%Y%j%H%M%S}.hdf"

@pytest.fixture(scope="session")
def modis_l2_nasa_mod35_file(tmpdir_factory) -> list[str]:
    """Create a single MOD35 L2 HDF4 file with headers."""
    filename = generate_nasa_l2_filename("MOD35")
    full_path = str(tmpdir_factory.mktemp("modis_l2").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 5000, include_angles=True)
    variable_infos.update(_get_cloud_mask_variable_info("Cloud_Mask", 1000))
    create_hdfeos_test_file(full_path,
                            variable_infos,
                            _create_struct_metadata(5000),
                            _create_core_metadata("MOD35"),
                            _create_header_metadata())
    return [full_path]

@pytest.fixture(scope="session")
def modis_l3_nasa_mcd12q1_file(tmpdir_factory) -> list[str]:
    """Create a single MOD35 L2 HDF4 file with headers."""
    filename = generate_nasa_l3_tile_filename("MCD12Q1")
    full_path = str(tmpdir_factory.mktemp("modis_l2").join(filename))
    variable_infos = _get_l3_land_cover_info()
    archive_header = \
"""GROUP                  = ARCHIVEDMETADATA
  GROUPTYPE            = MASTERGROUP

  OBJECT                 = NADIRDATARESOLUTION
    NUM_VAL              = 1
    VALUE                = "500m"
  END_OBJECT             = NADIRDATARESOLUTION

END_GROUP              = ARCHIVEDMETADATA

END
"""
    create_hdfeos_test_file(full_path,
                            variable_infos,
                            _create_struct_metadata_cmg("MCD12Q1"),
                            _create_core_metadata("MCD12Q1"),
                            archive_header)
    return [full_path]

def generate_nasa_l3_filename(prefix: str) -> str:
    """Generate a file name that follows MODIS 09 L3 convention in a temporary directory."""
    now = dt.datetime.now()
    return f"{prefix}.A{now:%Y%j}.061.{now:%Y%j%H%M%S}.hdf"


def modis_l3_file(tmpdir_factory, f_prefix, var_name, f_short):
    """Create a MODIS L3 file of the desired type."""
    filename = generate_nasa_l3_filename(f_prefix)
    full_path = str(tmpdir_factory.mktemp("modis_l3").join(filename))
    variable_infos = _get_l3_refl_variable_info(var_name)
    create_hdfeos_test_file(full_path,
                            variable_infos,
                            _create_struct_metadata_cmg(f_short),
                            _create_core_metadata(f_short),
                            _create_header_metadata())
    return [full_path]


@pytest.fixture(scope="session")
def modis_l3_nasa_mod09_file(tmpdir_factory) -> list[str]:
    """Create a single MOD09 L3 HDF4 file with headers."""
    return modis_l3_file(tmpdir_factory,
                         "MOD09CMG",
                         "Coarse_Resolution_Surface_Reflectance_Band_2",
                         "MOD09")


@pytest.fixture(scope="session")
def modis_l3_nasa_mod43_file(tmpdir_factory) -> list[str]:
    """Create a single MVCD43 L3 HDF4 file with headers."""
    return modis_l3_file(tmpdir_factory,
                         "MCD43C1",
                         "BRDF_Albedo_Parameter1_Band2",
                         "MCD43C1")


@pytest.fixture(scope="session")
def modis_l2_nasa_mod35_mod03_files(modis_l2_nasa_mod35_file, modis_l1b_nasa_mod03_file) -> list[str]:
    """Create a MOD35 L2 HDF4 file and MOD03 L1b geolocation file."""
    return modis_l2_nasa_mod35_file + modis_l1b_nasa_mod03_file


@pytest.fixture(scope="session")
def modis_l2_nasa_mod06_file(tmpdir_factory) -> list[str]:
    """Create a single MOD06 L2 HDF4 file with headers."""
    filename = generate_nasa_l2_filename("MOD06")
    full_path = str(tmpdir_factory.mktemp("modis_l2").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 5000, include_angles=True)
    variable_infos.update(_get_basic_variable_info("Surface_Pressure", 5000))
    variable_infos.update(_get_basic_variable_info("non_yaml_configured_2D_var", 5000))
    variable_infos.update(_get_basic_variable_info("non_yaml_configured_3D_var", 5000, dim_size=3))
    create_hdfeos_test_file(full_path,
                            variable_infos,
                            _create_struct_metadata(5000),
                            _create_core_metadata("MOD06"),
                            _create_header_metadata())
    return [full_path]

@pytest.fixture(scope="session")
def modis_l2_nasa_mod99_file(tmpdir_factory) -> list[str]:
    """Create an "artificial" MOD99 L2 HDF4 file with headers.

    There exists no MOD99 Level 2 product. This is just for testing available datasets
    in arbitrary level 2 file.
    """
    filename = generate_nasa_l2_filename("MOD99")
    full_path = str(tmpdir_factory.mktemp("modis_l2").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 5000, include_angles=True)
    variable_infos.update(_get_basic_variable_info("non_yaml_configured_2D_var", 1000))
    create_hdfeos_test_file(full_path,
                            variable_infos,
                            _create_struct_metadata(5000),
                            _create_core_metadata("MOD99"),
                            _create_header_metadata())
    return [full_path]

@pytest.fixture(scope="session")
def modis_l2_imapp_snowmask_file(tmpdir_factory) -> list[str]:
    """Create a single IMAPP snowmask L2 HDF4 file with headers."""
    filename = generate_imapp_filename("snowmask")
    full_path = str(tmpdir_factory.mktemp("modis_l2").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 5000, include_angles=False)
    variable_infos.update(_get_basic_variable_info("Snow_Mask", 1000))
    create_hdfeos_test_file(full_path, variable_infos)
    return [full_path]


@pytest.fixture(scope="session")
def modis_l2_imapp_snowmask_geo_files(modis_l2_imapp_snowmask_file, modis_l1b_nasa_mod03_file) -> list[str]:
    """Create the IMAPP snowmask and geo HDF4 files."""
    return modis_l2_imapp_snowmask_file + modis_l1b_nasa_mod03_file


@pytest.fixture(scope="session")
def modis_l2_imapp_mask_byte1_file(tmpdir_factory) -> list[str]:
    """Create a single IMAPP mask_byte1 L2 HDF4 file with headers."""
    filename = generate_imapp_filename("mask_byte1")
    full_path = str(tmpdir_factory.mktemp("modis_l2").join(filename))
    variable_infos = _get_l1b_geo_variable_info(filename, 5000, include_angles=False)
    variable_infos.update(_get_mask_byte1_variable_info())
    create_hdfeos_test_file(full_path, variable_infos)
    return [full_path]


@pytest.fixture(scope="session")
def modis_l2_imapp_mask_byte1_geo_files(modis_l2_imapp_mask_byte1_file, modis_l1b_nasa_mod03_file) -> list[str]:
    """Create the IMAPP mask_byte1 and geo HDF4 files."""
    return modis_l2_imapp_mask_byte1_file + modis_l1b_nasa_mod03_file



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2020, 2021, 2022 Pytroll developers

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for the AAPP AMSU-B/MHS level-1c data.

https://nwp-saf.eumetsat.int/site/download/documentation/aapp/NWPSAF-MF-UD-003_Formats_v8.0.pdf

"""

import logging

import dask.array as da
import numpy as np

from satpy.readers.aapp_l1b import AAPPL1BaseFileHandler, create_xarray
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger(__name__)


CHUNK_SIZE = get_legacy_chunk_size()
LINE_CHUNK = CHUNK_SIZE ** 2 // 90

MHS_AMSUB_CHANNEL_NAMES = ["1", "2", "3", "4", "5"]
MHS_AMSUB_ANGLE_NAMES = ["sensor_zenith_angle", "sensor_azimuth_angle",
                         "solar_zenith_angle", "solar_azimuth_difference_angle"]

MHS_AMSUB_PLATFORM_IDS2NAMES = {15: "NOAA-15",
                                16: "NOAA-16",
                                17: "NOAA-17",
                                18: "NOAA-18",
                                19: "NOAA-19",
                                1: "Metop-B",
                                2: "Metop-A",
                                3: "Metop-C",
                                4: "Metop simulator"}

MHS_AMSUB_PLATFORMS = ["Metop-A", "Metop-B", "Metop-C", "NOAA-18", "NOAA-19"]


class MHS_AMSUB_AAPPL1CFile(AAPPL1BaseFileHandler):
    """Reader for AMSU-B/MHS L1C files created from the AAPP software."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize object information by reading the input file."""
        super().__init__(filename, filename_info, filetype_info)

        self.channels = {i: None for i in MHS_AMSUB_CHANNEL_NAMES}
        self.units = {i: "brightness_temperature" for i in MHS_AMSUB_CHANNEL_NAMES}

        self._channel_names = MHS_AMSUB_CHANNEL_NAMES
        self._angle_names = MHS_AMSUB_ANGLE_NAMES

        self._set_filedata_layout()
        self.read()

        self._get_platform_name(MHS_AMSUB_PLATFORM_IDS2NAMES)
        self._get_sensorname()

    def _set_filedata_layout(self):
        """Set the file data type/layout."""
        self._header_offset = HEADER_LENGTH
        self._scan_type = _SCANTYPE
        self._header_type = _HEADERTYPE

    def _get_sensorname(self):
        """Get the sensor name from the header."""
        if self._header["instrument"][0] == 11:
            self.sensor = "amsub"
        elif self._header["instrument"][0] == 12:
            self.sensor = "mhs"
        else:
            raise IOError("Sensor neither MHS nor AMSU-B!")

    def get_angles(self, angle_id):
        """Get sun-satellite viewing angles."""
        satz = self._data["angles"][:, :, 0] * 1e-2
        sata = self._data["angles"][:, :, 1] * 1e-2

        sunz = self._data["angles"][:, :, 2] * 1e-2
        suna = self._data["angles"][:, :, 3] * 1e-2

        name_to_variable = dict(zip(MHS_AMSUB_ANGLE_NAMES, (satz, sata, sunz, suna)))
        return create_xarray(name_to_variable[angle_id])

    def navigate(self, coordinate_id):
        """Get the longitudes and latitudes of the scene."""
        lons, lats = self._get_coordinates_in_degrees()
        if coordinate_id == "longitude":
            return create_xarray(lons)
        if coordinate_id == "latitude":
            return create_xarray(lats)

        raise KeyError("Coordinate {} unknown.".format(coordinate_id))

    def _get_coordinates_in_degrees(self):
        lons = self._data["latlon"][:, :, 1] * 1e-4
        lats = self._data["latlon"][:, :, 0] * 1e-4
        return lons, lats

    def _calibrate_active_channel_data(self, key):
        """Calibrate active channel data only."""
        return self.calibrate(key)

    def calibrate(self, dataset_id):
        """Calibrate the data."""
        units = {"brightness_temperature": "K"}

        mask = True
        idx = ["1", "2", "3", "4", "5"].index(dataset_id["name"])

        ds = create_xarray(
            _calibrate(self._data, idx,
                       dataset_id["calibration"],
                       mask=mask))

        ds.attrs["units"] = units[dataset_id["calibration"]]
        ds.attrs.update(dataset_id._asdict())
        return ds


def _calibrate(data,
               chn,
               calib_type,
               mask=True):
    """Calibrate channel data.

    *calib_type* in brightness_temperature.

    """
    if calib_type not in ["brightness_temperature"]:
        raise ValueError("Calibration " + calib_type + " unknown!")

    channel = da.from_array(data["btemps"][:, :, chn] / 100., chunks=(LINE_CHUNK, 90))
    mask &= channel != 0

    if calib_type == "counts":
        return channel

    channel = channel.astype(np.float64)

    return da.where(mask, channel, np.nan)


HEADER_LENGTH = 1152*4

_HEADERTYPE = np.dtype([("siteid", "S3"),
                        ("cfill_1", "S1"),
                        ("l1bsite", "S3"),
                        ("cfill_2", "S1"),
                        ("versnb", "<i4"),
                        ("versyr", "<i4"),
                        ("versdy", "<i4"),
                        ("hdrcnt", "<i4"),
                        ("satid", "<i4"),
                        ("instrument", "<i4"),
                        ("satht", "<i4"),
                        ("period", "<i4"),
                        ("startorbit", "<i4"),
                        ("startdatayr", "<i4"),
                        ("startdatady", "<i4"),
                        ("startdatatime", "<i4"),
                        ("endorbit", "<i4"),
                        ("enddatayr", "<i4"),
                        ("enddatady", "<i4"),
                        ("enddatatime", "<i4"),
                        ("scnlin", "<i4"),
                        ("misscnlin", "<i4"),
                        ("vnantennacorr", "<i4"),
                        ("spare", "<i4"),
                        ("tempradcnv", "<i4", (3, 5)),
                        ("wmosatid", "<i4"),
                        ("filler", "<i4", (1114,)),
                        ])

_SCANTYPE = np.dtype([("scnlin", "<i4"),
                      ("scnlinyr", "<i4"),
                      ("scnlindy", "<i4"),
                      ("scnlintime", "<i4"),
                      ("qualind", "<i4"),
                      ("scnlinqual", "<i4"),
                      ("chanqual", "<i4", (5, )),
                      ("instrtemp", "<i4"),
                      ("spare1", "<i4", (2, )),
                      # Navigation
                      ("latlon", "<i4", (90, 2)),  # lat/lon in degrees for Bnfovs:
                      # first : 10^4 x (latitude)
                      # second : 10^4 x (longitude)
                      ("angles", "<i4", (90, 4)),  # scan angles for Bnfovs:
                      # first: 10^2 x (local zenith angle)
                      # second: 10^2 x (local azimuth angle)
                      # third: 10^2 x (solar zenith angle)
                      # fourth: 10^2 x (solar azimuth angle)
                      ("scalti", "<i4"),  # sat altitude above reference ellipsoid, km*10
                      ("spare2", "<i4", (2, )),
                      # Calibration
                      ("btemps", "<i4", (90, 5)),  # BT data for Bnfovs 10^2 x scene Tb (K), channels 1-5
                      ("dataqual", "<i4", (90, )),
                      ("filler", "<i4", (55, ))
                      ])



# Copyright (c) 2017-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Stretching."""

from __future__ import annotations

import logging
import typing
from collections import namedtuple
from numbers import Number
from typing import Optional

import dask.array as da
import numpy as np
import xarray as xr
from numpy.typing import ArrayLike

from .wrappers import exclude_alpha, using_map_blocks

if typing.TYPE_CHECKING:
    from trollimage.xrimage import XRImage

LOG = logging.getLogger(__name__)


def stretch(img, **kwargs):
    """Perform stretch."""
    return img.stretch(**kwargs)


def gamma(img, **kwargs):
    """Perform gamma correction."""
    return img.gamma(**kwargs)


def invert(img, *args):
    """Perform inversion."""
    return img.invert(*args)


def piecewise_linear_stretch(  # noqa: D417
        img: XRImage,
        xp: ArrayLike,
        fp: ArrayLike,
        reference_scale_factor: Optional[Number] = None,
        **kwargs) -> xr.DataArray:
    """Apply 1D linear interpolation.

    This uses :func:`numpy.interp` mapped over the provided dask array chunks.

    Args:
        img: Image data to be scaled. It is assumed the data is already
            normalized between 0 and 1.
        xp: Input reference values of the image data points used for
            interpolation. This is passed directly to :func:`numpy.interp`.
        fp: Target reference values of the output image data points used for
            interpolation. This is passed directly to :func:`numpy.interp`.
        reference_scale_factor: Divide ``xp`` and ``fp`` by this value before
            using them for interpolation. This is a convenience to make
            matching normalized image data to interp coordinates or to avoid
            floating point precision errors in YAML configuration files.
            If not provided, ``xp`` and ``fp`` will not be modified.

    Examples:
        This example YAML uses a 'crude' stretch to pre-scale the RGB data
        and then uses reference points in a 0-255 range.

        .. code-block:: yaml

              true_color_linear_interpolation:
                sensor: abi
                standard_name: true_color
                operations:
                - name: reflectance_range
                  method: !!python/name:satpy.enhancements.stretching.stretch
                  kwargs: {stretch: 'crude', min_stretch: 0., max_stretch: 100.}
                - name: Linear interpolation
                  method: !!python/name:satpy.enhancements.stretching.piecewise_linear_stretch
                  kwargs:
                   xp: [0., 25., 55., 100., 255.]
                   fp: [0., 90., 140., 175., 255.]
                   reference_scale_factor: 255

        This example YAML does the same as the above on the C02 channel, but
        the interpolation reference points are already adjusted for the input
        reflectance (%) data and the output range (0 to 1).

        .. code-block:: yaml

              c02_linear_interpolation:
                sensor: abi
                standard_name: C02
                operations:
                - name: Linear interpolation
                  method: !!python/name:satpy.enhancements.stretching.piecewise_linear_stretch
                  kwargs:
                   xp: [0., 9.8039, 21.5686, 39.2157, 100.]
                   fp: [0., 0.3529, 0.5490, 0.6863, 1.0]

    """
    LOG.debug("Applying the piecewise_linear_stretch")
    if reference_scale_factor is not None:
        xp = np.asarray(xp) / reference_scale_factor
        fp = np.asarray(fp) / reference_scale_factor

    return _piecewise_linear(img.data, xp=xp, fp=fp)


@exclude_alpha
@using_map_blocks
def _piecewise_linear(band_data, xp, fp):
    # Interpolate band on [0,1] using "lazy" arrays (put calculations off until the end).
    interp_data = np.interp(band_data, xp=xp, fp=fp)
    interp_data = np.clip(interp_data, 0, 1, out=interp_data)
    return interp_data


def cira_stretch(img, **kwargs):
    """Logarithmic stretch adapted to human vision.

    Applicable only for visible channels.
    """
    LOG.debug("Applying the cira-stretch")
    return _cira_stretch(img.data)


@exclude_alpha
def _cira_stretch(band_data):
    dtype = band_data.dtype
    log_root = np.log10(0.0223, dtype=dtype)
    denom = (1.0 - log_root) * 0.75
    band_data *= 0.01
    band_data = band_data.clip(np.finfo(float).eps)
    band_data = np.log10(band_data, dtype=dtype)
    band_data -= log_root
    band_data /= denom
    return band_data


def reinhard_to_srgb(img, saturation=1.25, white=100, **kwargs):  # noqa: D417
    """Stretch method based on the Reinhard algorithm, using luminance.

    Args:
        saturation: Saturation enhancement factor. Less is grayer. Neutral is 1.
        white: the reflectance luminance to set to white (in %).


    Reinhard, Erik & Stark, Michael & Shirley, Peter & Ferwerda, James. (2002).
    Photographic Tone Reproduction For Digital Images. ACM Transactions on Graphics.
    :doi: `21. 10.1145/566654.566575`
    """
    with xr.set_options(keep_attrs=True):
        # scale the data to [0, 1] interval
        rgb = img.data / 100
        white /= 100

        # extract color components
        r = rgb.sel(bands="R").data
        g = rgb.sel(bands="G").data
        b = rgb.sel(bands="B").data

        # saturate
        luma = _compute_luminance_from_rgb(r, g, b)
        rgb = (luma + (rgb - luma) * saturation).clip(0)

        # reinhard
        reinhard_luma = (luma / (1 + luma)) * (1 + luma / (white ** 2))
        coef = reinhard_luma / luma
        rgb = rgb * coef

        # srgb gamma
        rgb.data = _srgb_gamma(rgb.data)
        img.data = rgb

    return img.data


def _compute_luminance_from_rgb(r, g, b):
    """Compute the luminance of the image."""
    return r * 0.2126 + g * 0.7152 + b * 0.0722


def _srgb_gamma(arr):
    """Apply the srgb gamma."""
    return da.where(arr < 0.0031308, arr * 12.92, 1.055 * arr ** 0.41666 - 0.055)


def btemp_threshold(img, min_in, max_in, threshold, threshold_out=None, **kwargs):  # noqa: D417
    """Scale data linearly in two separate regions.

    This enhancement scales the input data linearly by splitting the data
    into two regions; min_in to threshold and threshold to max_in. These
    regions are mapped to 1 to threshold_out and threshold_out to 0
    respectively, resulting in the data being "flipped" around the
    threshold. A default threshold_out is set to `176.0 / 255.0` to
    match the behavior of the US National Weather Service's forecasting
    tool called AWIPS.

    Args:
        img (trollimage.xrimage.XRImage): Image object to be scaled
        min_in (float): Minimum input value to scale
        max_in (float): Maximum input value to scale
        threshold (float): Input value where to split data in to two regions
        threshold_out (float): Output value to map the input `threshold`
            to. Optional, defaults to 176.0 / 255.0.

    """
    threshold_out = threshold_out if threshold_out is not None else (176 / 255.0)
    low_factor = (threshold_out - 1.) / (min_in - threshold)
    low_offset = 1. + (low_factor * min_in)
    high_factor = threshold_out / (max_in - threshold)
    high_offset = high_factor * max_in

    Coeffs = namedtuple("Coeffs", "factor offset")
    high = Coeffs(high_factor, high_offset)
    low = Coeffs(low_factor, low_offset)

    return _bt_threshold(img.data,
                         threshold=threshold,
                         high_coeffs=high,
                         low_coeffs=low)


@exclude_alpha
@using_map_blocks
def _bt_threshold(band_data, threshold, high_coeffs, low_coeffs):
    # expects dask array to be passed
    return np.where(band_data >= threshold,
                    high_coeffs.offset - high_coeffs.factor * band_data,
                    low_coeffs.offset - low_coeffs.factor * band_data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Satpy developers
#
# This file is part of Satpy.
#
# Satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.smos_l2_wind module."""

import datetime as dt
import os
import unittest
from unittest import mock

import numpy as np
import xarray as xr

from satpy.tests.reader_tests.test_netcdf_utils import FakeNetCDF4FileHandler


class FakeNetCDF4FileHandlerSMOSL2WIND(FakeNetCDF4FileHandler):
    """Swap-in NetCDF4 File Handler."""

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        from xarray import DataArray
        dt_s = filename_info.get("start_time", dt.datetime(2020, 4, 22, 12, 0, 0))
        dt_e = filename_info.get("end_time", dt.datetime(2020, 4, 22, 12, 0, 0))

        if filetype_info["file_type"] == "smos_l2_wind":
            file_content = {
                "/attr/time_coverage_start": dt_s.strftime("%Y-%m-%dT%H:%M:%S Z"),
                "/attr/time_coverage_end": dt_e.strftime("%Y-%m-%dT%H:%M:%S Z"),
                "/attr/platform_shortname": "SM",
                "/attr/platform": "SMOS",
                "/attr/instrument": "MIRAS",
                "/attr/processing_level": "L2",
                "/attr/geospatial_bounds_vertical_crs": "EPSG:4623",
            }

            file_content["lat"] = np.arange(-90., 90.25, 0.25)
            file_content["lat/shape"] = (len(file_content["lat"]),)
            file_content["lat"] = DataArray(file_content["lat"], dims=("lat"))
            file_content["lat"].attrs["_FillValue"] = -999.0

            file_content["lon"] = np.arange(0., 360., 0.25)
            file_content["lon/shape"] = (len(file_content["lon"]),)
            file_content["lon"] = DataArray(file_content["lon"], dims=("lon"))
            file_content["lon"].attrs["_FillValue"] = -999.0

            file_content["wind_speed"] = np.ndarray(shape=(1,  # Time dimension
                                                           len(file_content["lat"]),
                                                           len(file_content["lon"])))
            file_content["wind_speed/shape"] = (1,
                                                len(file_content["lat"]),
                                                len(file_content["lon"]))
            file_content["wind_speed"] = DataArray(file_content["wind_speed"], dims=("time", "lat", "lon"),
                                                   coords=[[1], file_content["lat"], file_content["lon"]])
            file_content["wind_speed"].attrs["_FillValue"] = -999.0

        else:
            raise AssertionError()

        return file_content


class TestSMOSL2WINDReader(unittest.TestCase):
    """Test SMOS L2 WINDReader."""

    yaml_file = "smos_l2_wind.yaml"

    def setUp(self):
        """Wrap NetCDF4 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.smos_l2_wind import SMOSL2WINDFileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(SMOSL2WINDFileHandler, "__bases__", (FakeNetCDF4FileHandlerSMOSL2WIND,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the NetCDF4 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic initialization of this reader."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "SM_OPER_MIR_SCNFSW_20200420T021649_20200420T035013_110_001_7.nc",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    def test_load_wind_speed(self):
        """Load wind_speed dataset."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        with mock.patch("satpy.readers.smos_l2_wind.netCDF4.Variable", xr.DataArray):
            loadables = r.select_files_from_pathnames([
                "SM_OPER_MIR_SCNFSW_20200420T021649_20200420T035013_110_001_7.nc",
            ])
            r.create_filehandlers(loadables)
        ds = r.load(["wind_speed"])
        assert len(ds) == 1
        for d in ds.values():
            assert d.attrs["platform_shortname"] == "SM"
            assert d.attrs["sensor"] == "MIRAS"
            assert "area" in d.attrs
            assert d.attrs["area"] is not None
            assert "y" in d.dims
            assert "x" in d.dims
            assert d.shape == (719, 1440)
            assert d.y[0].data == -89.75
            assert d.y[d.shape[0] - 1].data == 89.75

    def test_load_lat(self):
        """Load lat dataset."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        with mock.patch("satpy.readers.smos_l2_wind.netCDF4.Variable", xr.DataArray):
            loadables = r.select_files_from_pathnames([
                "SM_OPER_MIR_SCNFSW_20200420T021649_20200420T035013_110_001_7.nc",
            ])
            r.create_filehandlers(loadables)
        ds = r.load(["lat"])
        assert len(ds) == 1
        for d in ds.values():
            assert "y" in d.dims
            assert d.shape == (719,)
            assert d.data[0] == -89.75
            assert d.data[d.shape[0] - 1] == 89.75

    def test_load_lon(self):
        """Load lon dataset."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        with mock.patch("satpy.readers.smos_l2_wind.netCDF4.Variable", xr.DataArray):
            loadables = r.select_files_from_pathnames([
                "SM_OPER_MIR_SCNFSW_20200420T021649_20200420T035013_110_001_7.nc",
            ])
            r.create_filehandlers(loadables)
        ds = r.load(["lon"])
        assert len(ds) == 1
        for d in ds.values():
            assert "x" in d.dims
            assert d.shape == (1440,)
            assert d.data[0] == -180.0
            assert d.data[d.shape[0] - 1] == 179.75

    def test_adjust_lon(self):
        """Load adjust longitude dataset."""
        from xarray import DataArray

        from satpy.readers.smos_l2_wind import SMOSL2WINDFileHandler
        smos_l2_wind_fh = SMOSL2WINDFileHandler("SM_OPER_MIR_SCNFSW_20200420T021649_20200420T035013_110_001_7.nc",
                                                {}, filetype_info={"file_type": "smos_l2_wind"})
        data = DataArray(np.arange(0., 360., 0.25), dims=("lon"))
        adjusted = smos_l2_wind_fh._adjust_lon_coord(data)
        expected = DataArray(np.concatenate((np.arange(0, 180., 0.25),
                                             np.arange(-180.0, 0, 0.25))),
                             dims=("lon"))
        assert adjusted.data.tolist() == expected.data.tolist()

    def test_roll_dataset(self):
        """Load roll of dataset along the lon coordinate."""
        from xarray import DataArray

        from satpy.readers.smos_l2_wind import SMOSL2WINDFileHandler
        smos_l2_wind_fh = SMOSL2WINDFileHandler("SM_OPER_MIR_SCNFSW_20200420T021649_20200420T035013_110_001_7.nc",
                                                {}, filetype_info={"file_type": "smos_l2_wind"})
        data = DataArray(np.arange(0., 360., 0.25), dims=("lon"))
        data = smos_l2_wind_fh._adjust_lon_coord(data)
        adjusted = smos_l2_wind_fh._roll_dataset_lon_coord(data)
        expected = np.arange(-180., 180., 0.25)
        assert adjusted.data.tolist() == expected.tolist()



# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Compositors using auxiliary data."""

from __future__ import annotations

import logging
import os

import satpy
from satpy.aux_download import DataDownloadMixin

from .core import GenericCompositor

LOG = logging.getLogger(__name__)


class StaticImageCompositor(GenericCompositor, DataDownloadMixin):
    """A compositor that loads a static image from disk.

    Environment variables in the filename are automatically expanded.

    """

    def __init__(self, name, filename=None, url=None, known_hash=None, area=None,  # noqa: D417
                 **kwargs):
        """Collect custom configuration values.

        Args:
            filename (str): Name to use when storing and referring to the file
                in the ``data_dir`` cache. If ``url`` is provided (preferred),
                then this is used as the filename in the cache and will be
                appended to ``<data_dir>/composites/<class_name>/``. If
                ``url`` is provided and ``filename`` is not then the
                ``filename`` will be guessed from the ``url``.
                If ``url`` is not provided, then it is assumed ``filename``
                refers to a local file. If the ``filename`` does not come with
                an absolute path, ``data_dir`` will be used as the directory path.
                Environment variables are expanded.
            url (str): URL to remote file. When the composite is created the
                file will be downloaded and cached in Satpy's ``data_dir``.
                Environment variables are expanded.
            known_hash (str or None): Hash of the remote file used to verify
                a successful download. If not provided then the download will
                not be verified. See :func:`satpy.aux_download.register_file`
                for more information.
            area (str): Name of area definition for the image.  Optional
                for images with built-in area definitions (geotiff).

        Use cases:
            1. url + no filename:
               Satpy determines the filename based on the filename in the URL,
               then downloads the URL, and saves it to <data_dir>/<filename>.
               If the file already exists and known_hash is also provided, then the pooch
               library compares the hash of the file to the known_hash. If it does not
               match, then the URL is re-downloaded. If it matches then no download.
            2. url + relative filename:
               Same as case 1 but filename is already provided so download goes to
               <data_dir>/<filename>. Same hashing behavior. This does not check for an
               absolute path.
            3. No url + absolute filename:
               No download, filename is passed directly to generic_image reader. No hashing
               is done.
            4. No url + relative filename:
               Check if <data_dir>/<filename> exists. If it does then make filename an
               absolute path. If it doesn't, then keep it as is and let the exception at
               the bottom of the method get raised.
        """
        filename, url = self._get_cache_filename_and_url(filename, url)
        self._cache_filename = filename
        self._url = url
        self._known_hash = known_hash
        self.area = None
        if area is not None:
            from satpy.area import get_area_def
            self.area = get_area_def(area)

        super(StaticImageCompositor, self).__init__(name, **kwargs)
        cache_keys = self.register_data_files([])
        self._cache_key = cache_keys[0]

    @staticmethod
    def _check_relative_filename(filename):
        data_dir = satpy.config.get("data_dir")
        path = os.path.join(data_dir, filename)

        return path if os.path.exists(path) else filename

    def _get_cache_filename_and_url(self, filename, url):
        filename = self._check_filename(filename, url)
        url, filename = self._check_url(url, filename)

        return filename, url

    def _check_filename(self, filename, url):
        if filename:
            filename = os.path.expanduser(os.path.expandvars(filename))
            if not os.path.isabs(filename) and not url:
                filename = self._check_relative_filename(filename)
        return filename

    def _check_url(self, url, filename):
        if url:
            url = os.path.expandvars(url)
            if not filename:
                filename = os.path.basename(url)
        elif not filename or not os.path.isabs(filename):
            raise ValueError("StaticImageCompositor needs a remote 'url', "
                             "or absolute path to 'filename', "
                             "or an existing 'filename' relative to Satpy's 'data_dir'.")
        return url, filename

    def register_data_files(self, data_files):
        """Tell Satpy about files we may want to download."""
        if os.path.isabs(self._cache_filename):
            return [None]
        return super().register_data_files([{
            "url": self._url,
            "known_hash": self._known_hash,
            "filename": self._cache_filename,
        }])

    def _retrieve_data_file(self):
        from satpy.aux_download import retrieve
        if os.path.isabs(self._cache_filename):
            return self._cache_filename
        return retrieve(self._cache_key)

    def __call__(self, *args, **kwargs):
        """Call the compositor."""
        from satpy import Scene
        local_file = self._retrieve_data_file()
        scn = Scene(reader="generic_image", filenames=[local_file])
        scn.load(["image"])
        img = scn["image"]
        # use compositor parameters as extra metadata
        # most important: set 'name' of the image
        img.attrs.update(self.attrs)
        # Check for proper area definition.  Non-georeferenced images
        # do not have `area` in the attributes
        if "area" not in img.attrs:
            if self.area is None:
                raise AttributeError("Area definition needs to be configured")
            img.attrs["area"] = self.area
        img.attrs["sensor"] = None
        img.attrs["mode"] = "".join(img.bands.data)
        img.attrs.pop("modifiers", None)
        img.attrs.pop("calibration", None)

        return img



# Copyright (c) 2010-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for Scene creation."""

import os
import random
import string
from copy import deepcopy
from unittest import mock

import pytest

import satpy
from satpy import Scene
from satpy.tests.utils import FAKE_FILEHANDLER_END, FAKE_FILEHANDLER_START, spy_decorator

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - include_test_etc


@pytest.mark.usefixtures("include_test_etc")
class TestScene:
    """Test the scene class."""

    def test_init(self):
        """Test scene initialization."""
        with mock.patch("satpy.scene.Scene._create_reader_instances") as cri:
            cri.return_value = {}
            Scene(filenames=["bla"], reader="blo")
            cri.assert_called_once_with(filenames=["bla"], reader="blo",
                                        reader_kwargs=None)

    def test_init_str_filename(self):
        """Test initializing with a single string as filenames."""
        with pytest.raises(ValueError, match="'filenames' must be a list of files: .*"):
            Scene(reader="blo", filenames="test.nc")

    def test_start_end_times(self):
        """Test start and end times for a scene."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        assert scene.start_time == FAKE_FILEHANDLER_START
        assert scene.end_time == FAKE_FILEHANDLER_END

    def test_init_preserve_reader_kwargs(self):
        """Test that the initialization preserves the kwargs."""
        cri = spy_decorator(Scene._create_reader_instances)
        with mock.patch("satpy.scene.Scene._create_reader_instances", cri):
            reader_kwargs = {"calibration_type": "gsics"}
            scene = Scene(filenames=["fake1_1.txt"],
                          reader="fake1",
                          filter_parameters={"area": "euron1"},
                          reader_kwargs=reader_kwargs)
            assert reader_kwargs is not cri.mock.call_args[1]["reader_kwargs"]
            assert scene.start_time == FAKE_FILEHANDLER_START
            assert scene.end_time == FAKE_FILEHANDLER_END

    def test_init_alone(self):
        """Test simple initialization."""
        scn = Scene()
        assert not scn._readers, "Empty scene should not load any readers"

    def test_init_no_files(self):
        """Test that providing an empty list of filenames fails."""
        with pytest.raises(ValueError, match="'filenames' was provided but is empty."):
            Scene(reader="viirs_sdr", filenames=[])

    def test_create_reader_instances_with_filenames(self):
        """Test creating a reader providing filenames."""
        filenames = ["bla", "foo", "bar"]
        reader_name = None
        with mock.patch("satpy.scene.load_readers") as findermock:
            Scene(filenames=filenames)
            findermock.assert_called_once_with(
                filenames=filenames,
                reader=reader_name,
                reader_kwargs=None,
            )

    def test_init_with_empty_filenames(self):
        """Test initialization with empty filename list."""
        filenames = []
        Scene(filenames=filenames)

    def test_init_with_fsfile(self):
        """Test initialisation with FSFile objects."""
        from satpy.readers.core.remote import FSFile

        # We should not mock _create_reader_instances here, because in
        # https://github.com/pytroll/satpy/issues/1605 satpy fails with
        # TypeError within that method if passed an FSFile instance.
        # Instead rely on the ValueError that satpy raises if no readers
        # are found.
        # Choose random filename that doesn't exist.  Not using tempfile here,
        # because tempfile creates files and we don't want that here.
        fsf = FSFile("".join(random.choices(string.printable, k=50)))
        with pytest.raises(ValueError, match="No supported files found"):
            Scene(filenames=[fsf], reader=[])

    def test_create_reader_instances_with_reader(self):
        """Test createring a reader instance providing the reader name."""
        reader = "foo"
        filenames = ["1", "2", "3"]
        with mock.patch("satpy.scene.load_readers") as findermock:
            findermock.return_value = {}
            Scene(reader=reader, filenames=filenames)
            findermock.assert_called_once_with(reader=reader,
                                               filenames=filenames,
                                               reader_kwargs=None,
                                               )

    def test_create_reader_instances_with_reader_kwargs(self):
        """Test creating a reader instance with reader kwargs."""
        from satpy.readers.core.yaml_reader import FileYAMLReader
        reader_kwargs = {"calibration_type": "gsics"}
        filter_parameters = {"area": "euron1"}
        reader_kwargs2 = {"calibration_type": "gsics", "filter_parameters": filter_parameters}

        rinit = spy_decorator(FileYAMLReader.create_filehandlers)
        with mock.patch("satpy.readers.core.yaml_reader.FileYAMLReader.create_filehandlers", rinit):
            scene = Scene(filenames=["fake1_1.txt"],
                          reader="fake1",
                          filter_parameters={"area": "euron1"},
                          reader_kwargs=reader_kwargs)
            del scene
            assert reader_kwargs == rinit.mock.call_args[1]["fh_kwargs"]
            rinit.mock.reset_mock()
            scene = Scene(filenames=["fake1_1.txt"],
                          reader="fake1",
                          reader_kwargs=reader_kwargs2)
            assert reader_kwargs == rinit.mock.call_args[1]["fh_kwargs"]
            del scene

    def test_create_multiple_reader_different_kwargs(self, include_test_etc):
        """Test passing different kwargs to different readers."""
        from satpy.readers.core.loading import load_reader
        with mock.patch.object(satpy.readers.core.loading, "load_reader", wraps=load_reader) as lr:
            Scene(filenames={"fake1_1ds": ["fake1_1ds_1.txt"],
                             "fake2_1ds": ["fake2_1ds_1.txt"]},
                  reader_kwargs={
                      "fake1_1ds": {"mouth": "omegna"},
                      "fake2_1ds": {"mouth": "varallo"}
                  })
            lr.assert_has_calls([
                mock.call([os.path.join(include_test_etc, "readers", "fake1_1ds.yaml")], mouth="omegna"),
                mock.call([os.path.join(include_test_etc, "readers", "fake2_1ds.yaml")], mouth="varallo")])

    def test_storage_options_from_reader_kwargs_no_options(self):
        """Test getting storage options from reader kwargs.

        Case where there are no options given.
        """
        filenames = ["s3://data-bucket/file1", "s3://data-bucket/file2", "s3://data-bucket/file3"]
        with mock.patch("satpy.scene.load_readers"):
            with mock.patch("fsspec.open_files") as open_files:
                Scene(filenames=filenames)
                open_files.assert_called_once_with(filenames)

    def test_storage_options_from_reader_kwargs_single_dict_no_options(self):
        """Test getting storage options from reader kwargs for remote files.

        Case where a single dict is given for all readers without storage options.
        """
        filenames = ["s3://data-bucket/file1", "s3://data-bucket/file2", "s3://data-bucket/file3"]
        reader_kwargs = {"reader_opt": "foo"}
        with mock.patch("satpy.scene.load_readers"):
            with mock.patch("fsspec.open_files") as open_files:
                Scene(filenames=filenames, reader_kwargs=reader_kwargs)
                open_files.assert_called_once_with(filenames)

    @pytest.mark.parametrize("reader_kwargs", [{}, {"reader_opt": "foo"}])
    def test_storage_options_from_reader_kwargs_single_dict(self, reader_kwargs):
        """Test getting storage options from reader kwargs.

        Case where a single dict is given for all readers with some common storage options.
        """
        filenames = ["s3://data-bucket/file1", "s3://data-bucket/file2", "s3://data-bucket/file3"]
        expected_reader_kwargs = reader_kwargs.copy()
        storage_options = {"option1": "1"}
        reader_kwargs["storage_options"] = storage_options
        orig_reader_kwargs = deepcopy(reader_kwargs)
        with mock.patch("satpy.scene.load_readers") as load_readers:
            with mock.patch("fsspec.open_files") as open_files:
                Scene(filenames=filenames, reader_kwargs=reader_kwargs)
                call_ = load_readers.mock_calls[0]
                assert call_.kwargs["reader_kwargs"] == expected_reader_kwargs
                open_files.assert_called_once_with(filenames, **storage_options)
                assert reader_kwargs == orig_reader_kwargs

    def test_storage_options_from_reader_kwargs_per_reader(self):
        """Test getting storage options from reader kwargs.

        Case where each reader have their own storage options.
        """
        filenames = {
            "reader1": ["s3://data-bucket/file1"],
            "reader2": ["s3://data-bucket/file2"],
            "reader3": ["s3://data-bucket/file3"],
        }
        storage_options_1 = {"option1": "1"}
        storage_options_2 = {"option2": "2"}
        storage_options_3 = {"option3": "3"}
        reader_kwargs = {
            "reader1": {"reader_opt_1": "foo"},
            "reader2": {"reader_opt_2": "bar"},
            "reader3": {"reader_opt_3": "baz"},
        }
        expected_reader_kwargs = deepcopy(reader_kwargs)
        reader_kwargs["reader1"]["storage_options"] = storage_options_1
        reader_kwargs["reader2"]["storage_options"] = storage_options_2
        reader_kwargs["reader3"]["storage_options"] = storage_options_3
        orig_reader_kwargs = deepcopy(reader_kwargs)

        with mock.patch("satpy.scene.load_readers") as load_readers:
            with mock.patch("fsspec.open_files") as open_files:
                Scene(filenames=filenames, reader_kwargs=reader_kwargs)
                call_ = load_readers.mock_calls[0]
                assert call_.kwargs["reader_kwargs"] == expected_reader_kwargs
                assert mock.call(filenames["reader1"], **storage_options_1) in open_files.mock_calls
                assert mock.call(filenames["reader2"], **storage_options_2) in open_files.mock_calls
                assert mock.call(filenames["reader3"], **storage_options_3) in open_files.mock_calls
                assert reader_kwargs == orig_reader_kwargs

    def test_storage_options_from_reader_kwargs_per_reader_and_global(self):
        """Test getting storage options from reader kwargs.

        Case where each reader have their own storage options and there are
        global options to merge.
        """
        filenames = {
            "reader1": ["s3://data-bucket/file1"],
            "reader2": ["s3://data-bucket/file2"],
            "reader3": ["s3://data-bucket/file3"],
        }
        reader_kwargs = {
            "reader1": {"reader_opt_1": "foo", "storage_options": {"option1": "1"}},
            "reader2": {"reader_opt_2": "bar", "storage_options": {"option2": "2"}},
            "storage_options": {"endpoint_url": "url"},
        }
        orig_reader_kwargs = deepcopy(reader_kwargs)

        with mock.patch("satpy.scene.load_readers"):
            with mock.patch("fsspec.open_files") as open_files:
                Scene(filenames=filenames, reader_kwargs=reader_kwargs)
                assert mock.call(filenames["reader1"], option1="1", endpoint_url="url") in open_files.mock_calls
                assert mock.call(filenames["reader2"], option2="2", endpoint_url="url") in open_files.mock_calls
                assert reader_kwargs == orig_reader_kwargs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2024 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Landsat reader.

Details of the data format can be found here:
  https://d9-wret.s3.us-west-2.amazonaws.com/assets/palladium/production/s3fs-public/media/files/LSDS-1822_Landsat8-9-OLI-TIRS_C2_L1_DataFormatControlBook-v7.pdf
  https://d9-wret.s3.us-west-2.amazonaws.com/assets/palladium/production/s3fs-public/atoms/files/LSDS-1414_Landsat7ETM-C2-L1-DFCB-v3.pdf
  https://d9-wret.s3.us-west-2.amazonaws.com/assets/palladium/production/s3fs-public/atoms/files/LSDS-1415_Landsat4-5-TM-C2-L1-DFCB-v3.pdf
  https://d9-wret.s3.us-west-2.amazonaws.com/assets/palladium/production/s3fs-public/atoms/files/LSDS-1416_LandsatMSS-C2-L1-DFCB-v3.pdf
  https://d9-wret.s3.us-west-2.amazonaws.com/assets/palladium/production/s3fs-public/media/files/LSDS-1328_Landsat8-9_OLI-TIRS-C2-L2_DFCB-v7.pdf
  https://d9-wret.s3.us-west-2.amazonaws.com/assets/palladium/production/s3fs-public/media/files/LSDS-1337_Landsat7ETM-C2-L2-DFCB-v6.pdf
  https://d9-wret.s3.us-west-2.amazonaws.com/assets/palladium/production/s3fs-public/atoms/files/LSDS-1336_Landsat4-5-TM-C2-L2-DFCB-v4.pdf
  https://d9-wret.s3.us-west-2.amazonaws.com/assets/palladium/production/s3fs-public/media/files/LSDS-1927_Landsat7-Data-Users-Handbook-v3.pdf
  https://www.usgs.gov/landsat-missions/using-usgs-landsat-level-1-data-product
  https://www.usgs.gov/landsat-missions/landsat-collection-2-level-2-science-products

NOTE: The scene geometry data (SZA, VZA, SAA, VAA) is retrieved from the L1 TIFF files, which are derived from Band 04.
The geometry differs between bands, so if you need precise geometry you should calculate this from the metadata instead.

"""

import logging
from datetime import datetime, timezone

import dask.array as da
import defusedxml.ElementTree as ET
import numpy as np
import rioxarray  # noqa: F401  # need by xarray with the engine rasterio
import xarray as xr

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.remote import open_file_or_filename

logger = logging.getLogger(__name__)

PLATFORMS = {"01": "Landsat-1",
             "02": "Landsat-2",
             "03": "Landsat-3",
             "04": "Landsat-4",
             "05": "Landsat-5",
             "07": "Landsat-7",
             "08": "Landsat-8",
             "09": "Landsat-9"}

PAN_BANDLIST = ["B8"]

ANGLIST = ["satellite_azimuth_angle",
           "satellite_zenith_angle",
           "solar_azimuth_angle",
           "solar_zenith_angle"]

ANGLIST_CHAN = ["sza", "saa", "vaa", "vza"]
QALIST_CHAN = ["qa", "qa_radsat", "qa_aerosol", "qa_atmos_opacity", "qa_cloud", "qa_st"]


class BaseLandsatReader(BaseFileHandler):
    """Basic file handler for Landsat files (tif)."""

    @staticmethod
    def get_btype(file_type):
        """Return the band type from the file type."""
        if "_B" in file_type:
            pos = file_type.rfind("_B")
        else:
            pos = file_type.rfind("_")
        if pos == -1:
            raise ValueError(f"Invalid file type: {file_type}")
        else:
            return file_type[pos+1:]

    @property
    def start_time(self):
        """Return start time."""
        return self._mda.start_time

    @property
    def end_time(self):
        """Return end time."""
        return self._mda.end_time

    @property
    def spectral_bands(self):
        """Landsat bands."""
        raise NotImplementedError()

    @property
    def thermal_bands(self):
        """Landsat thermal bands."""
        raise NotImplementedError()

    @property
    def sensor(self):
        """Sensor name."""
        raise NotImplementedError()

    def __init__(self, filename, filename_info, filetype_info, mda, **kwargs):
        """Initialize the reader."""
        super().__init__(filename, filename_info, filetype_info)

        # Check we have landsat data
        if filename_info["platform_type"] != "L":
            raise ValueError("This reader only supports Landsat data")

        # Get the channel name
        self.channel = self.get_btype(filetype_info["file_type"])

        # Data can be VIS, TIR or Combined. This flag denotes what the granule contains (O, T or C respectively).
        self.chan_selector = filename_info["data_type"]

        self._obs_date = filename_info["observation_date"]
        self._mda = mda

        # Retrieve some per-band useful metadata
        self.bsat = self._mda.band_saturation
        self.calinfo = self._mda.band_calibration
        self.platform_name = PLATFORMS[filename_info["spacecraft_id"]]
        self.collection = filename_info["collection_id"]

    def get_dataset(self, key, info):
        """Load a dataset."""
        self._check_channel_availability(key["name"])

        logger.debug("Reading %s.", key["name"])

        # xarray use the engine 'rasterio' to open the file, but
        #   its actually rioxarray used in the backend.
        #   however, error is not explicit enough (see https://github.com/pydata/xarray/issues/7831)
        data = xr.open_dataarray(open_file_or_filename(self.filename), engine="rasterio",
                                 chunks={"band": 1,
                                         "y": "auto",
                                         "x": "auto"},
                                 mask_and_scale=False).squeeze()

        self._mask_data(data, key["name"])

        attrs = self._collect_attrs(data, key["name"], info)

        # Rename to Satpy convention
        data = data.rename({"band": "bands"})

        data.attrs.update(attrs)

        # Calibrate if we're using a band rather than a QA or geometry dataset
        if "calibration" in key:
            data = self.calibrate(data, key["calibration"])

        self._scale_angle_data(data, key["name"])

        return data

    def _check_channel_availability(self, name):
        self._check_channel_and_reader(name)
        # OLI-TIRS sensor data sometimes can contain only OLI or only TIRS data
        self._check_oli_tirs_spectral(name)
        self._check_oli_tirs_thermal(name)

    def _check_channel_and_reader(self, name):
        if self.channel == name:
            return
        if self.channel in ANGLIST_CHAN:
            return
        if name in QALIST_CHAN:
            return
        raise ValueError(f"Requested channel {name} does not match the reader channel {self.channel}")

    def _check_oli_tirs_spectral(self, name):
        if self.sensor != "OLI_TIRS":
            return
        if name not in self.spectral_bands:
            return
        if self.chan_selector in ["O", "C"]:
            return
        raise ValueError(f"Requested channel {name} is not available in this granule")

    def _check_oli_tirs_thermal(self, name):
        if self.sensor != "OLI_TIRS":
            return
        if name not in self.thermal_bands:
            return
        if self.chan_selector in ["T", "C"]:
            return
        raise ValueError(f"Requested channel {name} is not available in this granule")

    def _mask_data(self, data, name):
        # For calibration simplicity convert the fill value to np.nan
        bands_9999 = ["TRAD", "DRAD", "URAD", "ATRAN", "EMIS", "EMSD", "CDIST", "qa_st", "qa_atmos_opacity"]
        if self.collection == "02" and name in bands_9999:
            # The fill value for several Landsat L2 bands is '-9999'
            data.data = da.where(data.data == -9999, np.float32(np.nan), data.data)
        else:
            # The fill value for Landsat is '0'
            data.data = da.where(data.data == 0, np.float32(np.nan), data.data)

    def _collect_attrs(self, data, name, info):
        attrs = data.attrs.copy()
        # Add useful metadata to the attributes.
        attrs["perc_cloud_cover"] = self._mda.cloud_cover
        # Add platform / sensor attributes
        attrs["platform_name"] = self.platform_name
        attrs["sensor"] = self.sensor
        # Apply attrs from YAML
        if "standard_name" in info:
            attrs["standard_name"] = info["standard_name"]
        if "units" in info:
            attrs["units"] = info["units"]

        # Only OLI bands have a saturation flag
        if name in self.bsat:
            attrs["saturated"] = self.bsat[name]

        return attrs

    def _scale_angle_data(self, data, name):
        if name in ANGLIST:
            data.data = data.data * 0.01

    def calibrate(self, data, calibration):
        """Calibrate the data from counts into the desired units."""
        raise NotImplementedError()

    def get_area_def(self, dsid):
        """Get area definition of the image from the metadata."""
        return self._mda.build_area_def(dsid["name"])


class BaseLandsatL1Reader(BaseLandsatReader):
    """Basic file handler for Landsat L1 files (tif)."""

    def calibrate(self, data, calibration):
        """Calibrate the data from counts into the desired units."""
        if calibration == "counts":
            return data

        if calibration in ["radiance", "brightness_temperature"]:
            data.data = data.data * self.calinfo[self.channel][0] + self.calinfo[self.channel][1]
            if calibration == "radiance":
                data.data = data.data.astype(np.float32)
                return data

        if calibration == "reflectance":
            data.data = data.data * self.calinfo[self.channel][2] + self.calinfo[self.channel][3]
            data.data = data.data.astype(np.float32) * 100
            return data

        if calibration == "brightness_temperature":
            data.data = (self.calinfo[self.channel][3] / np.log((self.calinfo[self.channel][2] / data.data) + 1))
            data.data = data.data.astype(np.float32)
            return data


class BaseLandsatL2Reader(BaseLandsatReader):
    """Basic file handler for Landsat L2 files (tif)."""

    def calibrate(self, data, calibration):
        """Calibrate the data from counts into the desired units."""
        if calibration == "counts":
            return data

        if calibration in ["reflectance", "brightness_temperature"]:
            data.data = data.data * self.calinfo[self.channel][0] + self.calinfo[self.channel][1]
            if calibration == "reflectance":
                data.data = data.data * 100
            data.data = data.data.astype(np.float32)
            return data

        if calibration == "radiance":
            data.data = data.data * 0.001
            data.data = data.data.astype(np.float32)
            return data

        if calibration in ["emissivity", "atmospheric_transmittance"]:
            data.data = data.data * 0.0001
            data.data = data.data.astype(np.float32)
            return data

        if calibration in ["qa_brightness_temperature", "cloud_distance"]:
            data.data = data.data * 0.01
            data.data = data.data.astype(np.float32)
            return data


class OLITIRSCHReader(BaseLandsatL1Reader):
    """File handler for Landsat OLI-TIRS L1 files (tif)."""

    @property
    def spectral_bands(self):
        """Landsat bands."""
        return ["B1", "B2", "B3", "B4", "B5", "B6", "B7", "B8", "B9"]

    @property
    def thermal_bands(self):
        """Landsat thermal bands."""
        return ["B10", "B11"]

    @property
    def sensor(self):
        """Sensor name."""
        return "OLI_TIRS"


class OLITIRSL2CHReader(BaseLandsatL2Reader):
    """File handler for Landsat OLI-TIRS L2 files (tif)."""

    @property
    def spectral_bands(self):
        """Landsat bands."""
        return ["B1", "B2", "B3", "B4", "B5", "B6", "B7"]

    @property
    def thermal_bands(self):
        """Landsat thermal bands."""
        return ["B10"]

    @property
    def sensor(self):
        """Sensor name."""
        return "OLI_TIRS"


class ETMCHReader(BaseLandsatL1Reader):
    """File handler for Landsat ETM+ L1 files (tif)."""

    @property
    def spectral_bands(self):
        """Landsat bands."""
        return ["B1", "B2", "B3", "B4", "B5", "B7", "B8"]

    @property
    def thermal_bands(self):
        """Landsat thermal bands."""
        return ["B6_VCID_1", "B6_VCID_2"]

    @property
    def sensor(self):
        """Sensor name."""
        return "ETM+"


class ETML2CHReader(BaseLandsatL2Reader):
    """File handler for Landsat ETM+ L2 files (tif)."""

    @property
    def spectral_bands(self):
        """Landsat bands."""
        return ["B1", "B2", "B3", "B4", "B5", "B7"]

    @property
    def thermal_bands(self):
        """Landsat thermal bands."""
        return ["B6"]

    @property
    def sensor(self):
        """Sensor name."""
        return "ETM+"


class TMCHReader(BaseLandsatL1Reader):
    """File handler for Landsat TM L1 files (tif)."""

    @property
    def spectral_bands(self):
        """Landsat bands."""
        return ["B1", "B2", "B3", "B4", "B5", "B7"]

    @property
    def thermal_bands(self):
        """Landsat thermal bands."""
        return ["B6"]

    @property
    def sensor(self):
        """Sensor name."""
        return "TM"


class TML2CHReader(BaseLandsatL2Reader):
    """File handler for Landsat TM L2 files (tif)."""

    @property
    def spectral_bands(self):
        """Landsat bands."""
        return ["B1", "B2", "B3", "B4", "B5", "B7"]

    @property
    def thermal_bands(self):
        """Landsat thermal bands."""
        return ["B6"]

    @property
    def sensor(self):
        """Sensor name."""
        return "TM"


class MSSCHReader(BaseLandsatL1Reader):
    """File handler for Landsat MSS L1 files (tif)."""

    @property
    def spectral_bands(self):
        """Landsat bands."""
        if self.platform_name in ["Landsat-1", "Landsat-2", "Landsat-3"]:
            return ["B4", "B5", "B6", "B7"]
        elif self.platform_name in ["Landsat-4", "Landsat-5"]:
            return ["B1", "B2", "B3", "B4"]

    @property
    def thermal_bands(self):
        """Landsat thermal bands."""
        return []

    @property
    def sensor(self):
        """Sensor name."""
        return "MSS"

    def available_datasets(self, configured_datasets=None):
        """Set up wavelength to B4 band."""
        # update previously configured datasets
        for is_avail, ds_info in (configured_datasets or []):
            availability, info = self._get_dataset_info(ds_info, is_avail)
            yield availability, info

    def _get_dataset_info(self, ds_info, is_avail):
        # some other file handler knows how to load this
        # don't override what they've done
        if is_avail is not None:
            return is_avail, ds_info

        matches = self.file_type_matches(ds_info["file_type"])
        if matches:
            info = self._get_matched_dataset_info(ds_info)
            return True, info

        # we don't know what to do with this
        # see if another future file handler does
        return is_avail, ds_info

    def _get_matched_dataset_info(self, ds_info):
        if ds_info.get("name") == "B4":
            return self._get_modified_wavelength_info(ds_info)
        return True, ds_info

    def _get_modified_wavelength_info(self, ds_info):
        # Modify the dataset's wavelength dynamically
        new_info = ds_info.copy()
        if self.platform_name in ["Landsat-1", "Landsat-2", "Landsat-3"]:
            new_info["wavelength"] = [0.5, 0.55, 0.6]  # Green
        elif self.platform_name in ["Landsat-4", "Landsat-5"]:
            new_info["wavelength"] = [0.8, 0.95, 1.1]  # NIR

        return new_info


class BaseLandsatMDReader(BaseFileHandler):
    """Metadata file handler for Landsat files (tif)."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init the reader."""
        super().__init__(filename, filename_info, filetype_info)
        # Check we have landsat data
        if filename_info["platform_type"] != "L":
            raise ValueError("This reader only supports Landsat data")
        self.platform_name = PLATFORMS[filename_info["spacecraft_id"]]
        self._obs_date = filename_info["observation_date"]
        self.root = ET.parse(open_file_or_filename(self.filename))
        self.process_level = filename_info["process_level_correction"]
        import bottleneck  # noqa
        import geotiepoints  # noqa

    def get_cal_params(self, top_key, key_1, key_2):
        """Read the requested calibration parameters."""
        gain_flags = {}
        add_flags = {}

        for elem in self.root.findall(f".//{top_key}/*"):
            if elem.tag.startswith(f"{key_1}_BAND_"):
                band_num = _parse_band_num(elem, key_1)
                gain_flags[band_num] = float(elem.text)

            if elem.tag.startswith(f"{key_2}_BAND_"):
                band_num = _parse_band_num(elem, key_2)
                add_flags[band_num] = float(elem.text)

        return {key: tuple([gain_flags[key], add_flags[key]]) for key in gain_flags}

    @property
    def center_time(self):
        """Return center time."""
        return datetime.strptime(self.root.find(".//IMAGE_ATTRIBUTES/SCENE_CENTER_TIME").text[:-2],
                                 "%H:%M:%S.%f").replace(tzinfo=timezone.utc)

    @property
    def start_time(self):
        """Return start time.

        This is actually the scene center time, as we don't have the start time.
        It is constructed from the observation date (from the filename) and the center time (from the metadata).
        """
        return datetime(self._obs_date.year, self._obs_date.month, self._obs_date.day,
                        self.center_time.hour, self.center_time.minute, self.center_time.second,
                        tzinfo=timezone.utc)

    @property
    def end_time(self):
        """Return end time.

        This is actually the scene center time, as we don't have the end time.
        It is constructed from the observation date (from the filename) and the center time (from the metadata).
        """
        return datetime(self._obs_date.year, self._obs_date.month, self._obs_date.day,
                        self.center_time.hour, self.center_time.minute, self.center_time.second,
                        tzinfo=timezone.utc)

    @property
    def cloud_cover(self):
        """Return estimated granule cloud cover percentage."""
        return float(self.root.find(".//IMAGE_ATTRIBUTES/CLOUD_COVER").text)

    @property
    def band_saturation(self):
        """Return per-band saturation flag."""
        flags = {}
        for elem in self.root.findall(".//IMAGE_ATTRIBUTES/*"):
            if elem.tag.startswith("SATURATION_BAND_"):
                band_num = elem.tag.replace("SATURATION_BAND_", "B")
                flags[band_num] = (elem.text == "Y")
        return flags

    @property
    def band_calibration(self):
        """Return per-band saturation flag."""
        raise NotImplementedError()

    def earth_sun_distance(self):
        """Return Earth-Sun distance."""
        return float(self.root.find(".//IMAGE_ATTRIBUTES/EARTH_SUN_DISTANCE").text)

    def build_area_def(self, bname):
        """Build area definition from metadata."""
        from pyresample.geometry import AreaDefinition

        # Here we assume that the thermal bands have the same resolution as the reflective bands,
        # with only the panchromatic band (b08) having a different resolution.
        if bname in PAN_BANDLIST and self.platform_name in ["Landsat-7", "Landsat-8", "Landsat-9"]:
            pixoff = float(self.root.find(".//PROJECTION_ATTRIBUTES/GRID_CELL_SIZE_PANCHROMATIC").text) / 2.
            x_size = float(self.root.find(".//PROJECTION_ATTRIBUTES/PANCHROMATIC_SAMPLES").text)
            y_size = float(self.root.find(".//PROJECTION_ATTRIBUTES/PANCHROMATIC_LINES").text)
        else:
            pixoff = float(self.root.find(".//PROJECTION_ATTRIBUTES/GRID_CELL_SIZE_REFLECTIVE").text) / 2.
            x_size = float(self.root.find(".//PROJECTION_ATTRIBUTES/REFLECTIVE_SAMPLES").text)
            y_size = float(self.root.find(".//PROJECTION_ATTRIBUTES/REFLECTIVE_LINES").text)

        # Get remaining geoinfo from file
        datum = self.root.find(".//PROJECTION_ATTRIBUTES/DATUM").text

        # Reading utm zone or get specific crs for arctic and antarctic
        if self.root.find(".//PROJECTION_ATTRIBUTES/UTM_ZONE") is not None:
            utm_zone = self.root.find(".//PROJECTION_ATTRIBUTES/UTM_ZONE").text
            pcs_id = f"{datum} / UTM zone {utm_zone}N"
            proj_code = f"EPSG:326{utm_zone.zfill(2)}"
        else:
            lat_ts = self.root.find(".//PROJECTION_ATTRIBUTES/TRUE_SCALE_LAT").text
            if lat_ts == "-71.00000":
                # Antarctic
                proj_code = "EPSG:3031"
            if lat_ts == "71.00000":
                # Arctic
                proj_code = "EPSG:3995"
            pcs_id = f"{datum} / EPSG: {proj_code[5:]}N"

        # We need to subtract / add half a pixel from the corner to get the correct extent (pixel centers)
        ext_p1 = float(self.root.find(".//PROJECTION_ATTRIBUTES/CORNER_UL_PROJECTION_X_PRODUCT").text) - pixoff
        ext_p2 = float(self.root.find(".//PROJECTION_ATTRIBUTES/CORNER_LR_PROJECTION_Y_PRODUCT").text) - pixoff
        ext_p3 = float(self.root.find(".//PROJECTION_ATTRIBUTES/CORNER_LR_PROJECTION_X_PRODUCT").text) + pixoff
        ext_p4 = float(self.root.find(".//PROJECTION_ATTRIBUTES/CORNER_UL_PROJECTION_Y_PRODUCT").text) + pixoff

        # Create area definition
        area_extent = (ext_p1, ext_p2, ext_p3, ext_p4)

        # Return the area extent
        return AreaDefinition(f"EPSG_{proj_code[5:]}", pcs_id, pcs_id, proj_code, x_size, y_size, area_extent)


def _parse_band_num(elem, key):
    band_num = elem.tag.replace(f"{key}_BAND_", "")
    if band_num.startswith("ST_"):
        band_num = band_num.replace("ST_", "")
    else:
        band_num = "B" + band_num
    return band_num


class LandsatL1MDReader(BaseLandsatMDReader):
    """Metadata file handler for Landsat L1 files (tif)."""

    @property
    def band_calibration(self):
        """Return per-band calibration parameters."""
        radcal = self.get_cal_params("LEVEL1_RADIOMETRIC_RESCALING", "RADIANCE_MULT", "RADIANCE_ADD")
        viscal = self.get_cal_params("LEVEL1_RADIOMETRIC_RESCALING", "REFLECTANCE_MULT", "REFLECTANCE_ADD")
        tircal = self.get_cal_params("LEVEL1_THERMAL_CONSTANTS", "K1_CONSTANT", "K2_CONSTANT")
        topcal = viscal | tircal
        return {key: tuple([*radcal[key], *topcal[key]]) for key in radcal}


class LandsatL2MDReader(BaseLandsatMDReader):
    """Metadata file handler for Landsat L2 files (tif)."""

    @property
    def band_calibration(self):
        """Return per-band calibration parameters."""
        viscal = self.get_cal_params("LEVEL2_SURFACE_REFLECTANCE_PARAMETERS", "REFLECTANCE_MULT", "REFLECTANCE_ADD")
        tircal = self.get_cal_params("LEVEL2_SURFACE_TEMPERATURE_PARAMETERS", "TEMPERATURE_MULT", "TEMPERATURE_ADD")
        return viscal | tircal



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the SEVIRI L1b HDF4 from ICARE reader."""
import os
import unittest
from unittest import mock

import dask.array as da
import numpy as np
import pytest

from satpy.readers.core.loading import load_reader
from satpy.tests.reader_tests.test_hdf4_utils import FakeHDF4FileHandler

DEFAULT_FILE_DTYPE = np.uint16
DEFAULT_FILE_SHAPE = (10, 300)
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                              dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)


class FakeHDF4FileHandler2(FakeHDF4FileHandler):
    """Swap in HDF4 file handler."""

    def get_test_content(self, filename, filename_info, filename_type):
        """Mimic reader input file content."""
        file_content = {}
        file_content["/attr/Nadir_Pixel_Size"] = 3000.
        file_content["/attr/Beginning_Acquisition_Date"] = "2004-12-29T12:15:00Z"
        file_content["/attr/End_Acquisition_Date"] = "2004-12-29T12:27:44Z"
        file_content["/attr/Geolocation"] = ("1.3642337E7", "1856.0", "1.3642337E7", "1856.0")
        file_content["/attr/Altitude"] = "42164.0"
        file_content["/attr/Geographic_Projection"] = "geos"
        file_content["/attr/Projection_Longitude"] = "0.0"
        file_content["/attr/Sub_Satellite_Longitude"] = "3.4"
        file_content["/attr/Sensors"] = "MSG1/SEVIRI"
        file_content["/attr/Zone"] = "G"
        file_content["/attr/_FillValue"] = 1
        file_content["/attr/scale_factor"] = 1.
        file_content["/attr/add_offset"] = 0.

        # test one IR and one VIS channel
        file_content["Normalized_Radiance"] = DEFAULT_FILE_DATA
        file_content["Normalized_Radiance/attr/_FillValue"] = 1
        file_content["Normalized_Radiance/attr/scale_factor"] = 1.
        file_content["Normalized_Radiance/attr/add_offset"] = 0.
        file_content["Normalized_Radiance/shape"] = DEFAULT_FILE_SHAPE

        file_content["Brightness_Temperature"] = DEFAULT_FILE_DATA
        file_content["Brightness_Temperature/attr/_FillValue"] = 1
        file_content["Brightness_Temperature/attr/scale_factor"] = 1.
        file_content["Brightness_Temperature/attr/add_offset"] = 0.
        file_content["Brightness_Temperature/shape"] = DEFAULT_FILE_SHAPE

        # convert to xarrays
        from xarray import DataArray
        for key, val in file_content.items():
            if isinstance(val, np.ndarray):
                attrs = {}
                for a in ["_FillValue", "scale_factor", "add_offset"]:
                    if key + "/attr/" + a in file_content:
                        attrs[a] = file_content[key + "/attr/" + a]
                file_content[key] = DataArray(da.from_array(val), dims=("x", "y"), attrs=attrs)
        return file_content


class TestSEVIRIICAREReader(unittest.TestCase):
    """Test SEVIRI L1b HDF4 from ICARE Reader."""

    yaml_file = "seviri_l1b_icare.yaml"

    def setUp(self):
        """Wrap HDF4 file handler with own fake file handler."""
        from satpy._config import config_search_paths
        from satpy.readers.seviri_l1b_icare import SEVIRI_ICARE
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        self.p = mock.patch.object(SEVIRI_ICARE, "__bases__", (FakeHDF4FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the HDF4 file handler."""
        self.p.stop()

    def compare_areas(self, v):
        """Compare produced AreaDefinition with expected."""
        test_area = {"area_id": "geosmsg",
                     "width": 10,
                     "height": 300,
                     "area_extent": (-5567248.2834071,
                                     -5570248.6866857,
                                     -5537244.2506213,
                                     -4670127.7031114)}
        assert v.attrs["area"].area_id == test_area["area_id"]
        assert v.attrs["area"].width == test_area["width"]
        assert v.attrs["area"].height == test_area["height"]
        np.testing.assert_almost_equal(v.attrs["area"].area_extent,
                                       test_area["area_extent"])

    def test_init(self):
        """Test basic init with no extra parameters."""
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "GEO_L1B-MSG1_2004-12-29T12-15-00_G_VIS08_V1-04.hdf",
            "GEO_L1B-MSG1_2004-12-29T12-15-00_G_IR108_V1-04.hdf"
        ])
        assert len(loadables) == 2
        r.create_filehandlers(loadables)
        assert r.file_handlers

    def test_load_dataset_vis(self):
        """Test loading all datasets from a full swath file."""
        import datetime as dt
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "GEO_L1B-MSG1_2004-12-29T12-15-00_G_VIS08_V1-04.hdf"
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["VIS008"])
        assert len(datasets) == 1
        for v in datasets.values():
            date = dt.datetime(2004, 12, 29, 12, 27, 44)
            assert v.attrs["end_time"] == date
            assert v.attrs["calibration"] == "reflectance"

    def test_load_dataset_ir(self):
        """Test loading all datasets from a full swath file."""
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "GEO_L1B-MSG1_2004-12-29T12-15-00_G_IR108_V1-04.hdf"
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["IR_108"])
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.attrs["calibration"] == "brightness_temperature"

    def test_area_def_lores(self):
        """Test loading all datasets from an area of interest file."""
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "GEO_L1B-MSG1_2004-12-29T12-15-00_G_VIS08_V1-04.hdf"
        ])
        r.create_filehandlers(loadables)
        ds = r.load(["VIS008"])
        self.compare_areas(ds["VIS008"])
        assert ds["VIS008"].attrs["area"].proj_id == "msg_lowres"

    def test_area_def_hires(self):
        """Test loading all datasets from an area of interest file."""
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "GEO_L1B-MSG1_2004-12-29T12-15-00_G_HRV_V1-04.hdf",
        ])
        r.create_filehandlers(loadables)
        ds = r.load(["HRV"])
        self.compare_areas(ds["HRV"])
        assert ds["HRV"].attrs["area"].proj_id == "msg_hires"

    def test_sensor_names(self):
        """Check satellite name conversion is correct, including error case."""
        file_data = FakeHDF4FileHandler2.get_test_content(mock.MagicMock(),
                                                          mock.MagicMock(),
                                                          mock.MagicMock(),
                                                          mock.MagicMock())
        sensor_list = {"Meteosat-08": "MSG1/SEVIRI",
                       "Meteosat-09": "MSG2/SEVIRI",
                       "Meteosat-10": "MSG3/SEVIRI",
                       "Meteosat-11": "MSG4/SEVIRI"}
        with mock.patch("satpy.tests.reader_tests.test_seviri_l1b_icare."
                        "FakeHDF4FileHandler2.get_test_content") as patched_func:
            def _run_target():
                patched_func.return_value = file_data
                return self.p.target(mock.MagicMock(),
                                     mock.MagicMock(),
                                     mock.MagicMock()).sensor_name

            for sat in sensor_list:
                file_data["/attr/Sensors"] = sensor_list[sat]
                plat, sens = _run_target()
                assert plat == sat

            file_data["/attr/Sensors"] = "BADSAT/NOSENSE"
            with pytest.raises(NameError):
                plat, sens = _run_target()

    def test_bad_bandname(self):
        """Check reader raises an error if a band bandname is passed."""
        with pytest.raises(NameError):
            self.p.target(mock.MagicMock(),
                          mock.MagicMock(),
                          mock.MagicMock())._get_dsname({"name": "badband"})

    def test_nocompute(self):
        """Test that dask does not compute anything in the reader itself."""
        import dask

        from satpy.tests.utils import CustomScheduler
        with dask.config.set(scheduler=CustomScheduler(max_computes=0)):
            r = load_reader(self.reader_configs)
            loadables = r.select_files_from_pathnames([
                "GEO_L1B-MSG1_2004-12-29T12-15-00_G_VIS08_V1-04.hdf"
            ])
            r.create_filehandlers(loadables)
            r.load(["VIS008"])
